# -*- coding: utf-8 -*-
"""Mutable model

.. moduleauthor:: Konrad Hinsen

This is the main in-memory representation of Mosaic data.  Like most
Python code, it uses mutable data structures, which are efficient but
can become inconsistent if not used carefully.

"""

#-----------------------------------------------------------------------------
#       Copyright (C) 2013 The Mosaic Development Team
#
#  Distributed under the terms of the BSD License.  The full license is in
#  the file LICENSE.txt, distributed as part of this software.
#-----------------------------------------------------------------------------

import copy
import itertools as IT
import types

import numpy as N

import mosaic.api as api
from mosaic.utility import AbstractFactory
from mosaic.utility import MethodRegister
from mosaic.utility import uint_for_max_value


class MosaicObject(object):

    def __repr__(self):
        return str(self)

    def __ne__(self, other):
        return not self.__eq__(other)


class AtomDescriptor(MosaicObject):

    def __new__(cls, name=""):
        try:
            return cls._instances[name]
        except KeyError:
            ob = super(AtomDescriptor, cls).__new__(cls)
            api.validate_type(name, str, "name")
            cls._validate_name(name)
            ob._name = name
            cls._instances[name] = ob
            return ob

    @classmethod
    def _validate_name(cls, name):
        pass

    def __copy__(self):
        return self

    def __deepcopy__(self, memo):
        return self

    def __eq__(self, other):
        return self is other

    def __hash__(self):
        return id(self)

    def __str__(self):
        return "%s('%s')" % (self.__class__.__name__, self._name)

    @property
    def name(self):
        return self._name


class Element(AtomDescriptor):

    _instances = {}

    @classmethod
    def _validate_name(cls, name):
        api.MosaicAtom.validate_element_name(name)

    def __copy__(self):
        return self

    def __deepcopy__(self, memo):
        return self

    def __str__(self):
        return "Element(%s)" % self._name

    @property
    def type(self):
        return "element"


class CGParticle(AtomDescriptor):

    _instances = {}

    def __str__(self):
        if self._name:
            return "CGParticle('%s')" % self._name
        else:
            return "CGParticle()"

    @property
    def type(self):
        return "cgparticle"


class Dummy(AtomDescriptor):

    _instances = {}

    def __str__(self):
        if self._name:
            return "Dummy('%s')" % self._name
        else:
            return "Dummy()"

    @property
    def type(self):
        return "dummy"


class Unknown(AtomDescriptor):

    _instances = {}

    def __str__(self):
        if self._name:
            return "Unknown('%s')" % self._name
        else:
            return "Unknown()"

    @property
    def type(self):
        return ""


class Path(tuple):

    def __str__(self):
        return '.'.join(self)


class MolecularStructureObject(MosaicObject):

    def __copy__(self):
        return copy.deepcopy(self)

    def has_ancestor(self, obj):
        return self is obj or \
               (self._parent is not None and self._parent.has_ancestor(obj))

    def toplevel_ancestor(self):
        if self._parent is None:
            return self
        else:
            return self._parent.toplevel_ancestor()

    def ancestors(self, target=None):
        if self._parent is target:
            return (self,)
        else:
            if self._parent is None:
                raise ValueError("%s is not an ancestor of %s"
                                 % (str(target), str(self)))
            return self._parent.ancestors(target) + (self,)

    def common_ancestor(self, objs):
        ancestors = [obj.ancestors() for obj in objs]
        common = None
        while all(ancestors):
            top = set(a[0] for a in ancestors)
            if len(top) == 1:
                common = top.pop()
                ancestors = [a[1:] for a in ancestors]
            else:
                return common

    def path_to(self, obj):
        return Path(a._label for a in obj.ancestors(self))

    def full_path(self):
        return Path(obj._label for obj in self.ancestors())


class Atom(MolecularStructureObject, api.MosaicAtom):

    def __init__(self, label, descriptor, nsites=1):
        self.label = label

        api.validate_type(descriptor, AtomDescriptor, "atom descriptor")
        api.validate_type(nsites, int, "number of sites")
        if nsites < 1:
            raise ValueError("number of sites must be >= 1")

        self._descriptor = descriptor
        self.nsites = nsites
        self._parent = None

    def __deepcopy__(self, memo):
        return Atom(self._label, self._descriptor, self.nsites)

    def __getitem__(self, item):
        assert isinstance(item, str)
        raise KeyError(item)

    def __eq__(self, other):
        return isinstance(other, Atom) \
               and self.type == other.type \
               and self.name == other.name \
               and self.label == other.label \
               and self.nsites == other.nsites

    def __hash__(self):
        return hash(self.name + self.label + str(self.nsites))

    def __str__(self):
        return "%s('%s')" % (self.__class__.__name__, self._label)

    @property
    def parent(self):
        return self._parent

    # Data Model API implementation
    @property
    def label(self):
        return self._label

    @label.setter
    def label(self, new_label):
        api.validate_label(new_label, "Atom.label")
        self._label = new_label

    @property
    def name(self):
        return self._descriptor.name

    @property
    def type(self):
        return self._descriptor.type

    @property
    def number_of_sites(self):
        return self.nsites


class Fragment(MolecularStructureObject, api.MosaicFragment):

    def __init__(self, label, species, fragments=[], atoms=[], bonds=[]):
        self.label = label
        self.species = species

        self._parent = None
        self._fragments = []
        self._atoms = []
        self._bonds = []
        self._attrs = {}

        self.add_fragments(fragments)
        self.add_atoms(atoms)
        self.add_bonds(bonds)

    def __str__(self):
        return "%s('%s', '%s')" % (self.__class__.__name__,
                                   self._label, self._species)

    def add_fragments(self, fragments):
        api.validate_sequence(fragments, Fragment, "fragments")
        new_attrs = set()
        for f in fragments:
            self._check_child_obj(f, new_attrs)
        for f in fragments:
            self._add_child_obj(f)
        self._fragments.extend(fragments)

    def add_atoms(self, atoms):
        api.validate_sequence(atoms, Atom, "atoms")
        new_attrs = set()
        for a in atoms:
            self._check_child_obj(a, new_attrs)
        for a in atoms:
            self._add_child_obj(a)
        self._atoms.extend(atoms)

    def add_bonds(self, bonds):
        api.validate_sequence(bonds, tuple, "bonds",
                              ((lambda p: len(p) == 3,
                                "must have length  3"),
                               (lambda p: (isinstance(p[0], str)
                                           or isinstance(p[0], Atom))
                                          and (isinstance(p[1], str)
                                               or isinstance(p[1], Atom)),
                                "elements must be strings or atoms"),
                               (lambda p: p[2] in self._bond_orders,
                                "bond order must be one of "
                                + str(self._bond_orders))))
        for a1, a2, order in bonds:
            atom1 = self._get_atom(a1)
            atom2 = self._get_atom(a2)
            if atom1 is atom2:
                raise ValueError("bond between %s and itself"
                                 % str(atom1.full_path()))
            p = self.common_ancestor((atom1, atom2))
            if p is self:
                self._bonds.append((atom1, atom2, order))
            else:
                raise ValueError("bond %s-%s must be defined in fragment %s"
                                 % (str(atom1.full_path()),
                                    str(atom2.full_path()),
                                    p._label))

    def _check_child_obj(self, obj, new_attrs):
        if obj._parent is not None:
            raise ValueError("%s is already part of another fragment (%s)"
                             % (str(obj), obj._parent._label))
        l = obj._label
        if l in self._attrs or l in new_attrs:
            raise ValueError("Label %s occurs more than once" % l)
        new_attrs.add(l)

    def _add_child_obj(self, obj):
        obj._parent = self
        self._attrs[obj._label] = obj

    def __deepcopy__(self, memo):
        return Fragment(self._label, self._species,
                        copy.deepcopy(self._fragments, memo),
                        copy.deepcopy(self._atoms, memo),
                        tuple((str(self.path_to(a1)),
                               str(self.path_to(a2)),
                               order)
                              for a1, a2, order in self._bonds))

    def eq_structure(self, other):
        def bond_path_set(frag):
            set(tuple(sorted((a1, a2)) + [order])
                for a1, a2, order in frag.bonds)
        return self.species == other.species \
               and len(self.fragments) == len(other.fragments) \
               and len(self.atoms) == len(other.atoms) \
               and len(self.bonds) == len(other.bonds) \
               and all(f1 == f2
                       for f1, f2 in zip(self.fragments, other.fragments)) \
               and all(a1 == a2 for a1, a2 in zip(self.atoms, other.atoms)) \
               and bond_path_set(self) == bond_path_set(other)

    def __eq__(self, other):
        return isinstance(other, Fragment) \
               and self.label == other.label \
               and self.eq_structure(other)

    def __hash__(self):
        return hash(self.label + self.species)

    def _get_atom(self, atom_spec):
        if isinstance(atom_spec, str):
            try:
                return self[atom_spec]
            except KeyError:
                raise ValueError("fragment %s does not contain atom %s"
                                 % (self._label, atom_spec))
        else:
            if not atom_spec.has_ancestor(self):
                raise ValueError("fragment %s does not contain atom %s"
                                 % (self._label, str(atom_spec)))
            return atom_spec

    def __getitem__(self, path):
        assert isinstance(path, str) or isinstance(path, Path)
        obj = self._attrs
        if isinstance(path, str):
            path = path.split('.')
        for item in path:
            obj = obj[item]
        return obj

    @property
    def parent(self):
        return self._parent

    # Data Model API implementation
    @property
    def label(self):
        return self._label

    @label.setter
    def label(self, new_label):
        api.validate_label(new_label, "Fragment.label")
        self._label = new_label

    @property
    def species(self):
        return self._species

    @species.setter
    def species(self, new_species):
        api.validate_label(new_species, "Fragment.species")
        self._species = new_species

    @property
    def is_polymer(self):
        return False

    @property
    def polymer_type(self):
        raise ValueError("Non-polymer fragment does not have a polymer type")

    @property
    def fragments(self):
        return self._fragments

    @property
    def atoms(self):
        return self._atoms

    @property
    def bonds(self):
        return tuple((str(self.path_to(a1)),
                      str(self.path_to(a2)),
                      order)
                     for a1, a2, order in self._bonds)


class Polymer(Fragment):

    def __init__(self, label, species, fragments, bonds, polymer_type=''):
        Fragment.__init__(self, label, species, fragments, (), bonds)
        self.polymer_type = polymer_type

    def __deepcopy__(self, memo):
        return Polymer(self._label, self._species,
                       copy.deepcopy(self._fragments, memo),
                       tuple((str(self.path_to(a1)),
                              str(self.path_to(a2)),
                              order)
                             for a1, a2, order in self._bonds),
                       self._polymer_type)

    def __getitem__(self, item):
        if isinstance(item, str):
            return Fragment.__getitem__(self, item)
        assert isinstance(item, int)
        return self._fragments[item]

    def __len__(self):
        return len(self._fragments)

    def __eq__(self, other):
        return isinstance(other, Polymer) \
               and Fragment.__eq__(self, other) \
               and self.polymer_type == other.polymer_type

    __hash__ = Fragment.__hash__

    # Data Model API implementation
    @property
    def is_polymer(self):
        return True

    @property
    def polymer_type(self):
        return self._polymer_type

    @polymer_type.setter
    def polymer_type(self, new_polymer_type):
        api.validate_value(new_polymer_type, self._polymer_types,
                           'polymer_type')
        self._polymer_type = new_polymer_type


class Universe(MosaicObject, api.MosaicUniverse):

    def __init__(self, cell_shape, molecules,
                 symmetry_transformations=(),
                 convention=''):
        self.cell_shape = cell_shape
        self.symmetry_transformations = symmetry_transformations
        self.convention = convention
        self._molecules = []
        self._templates = []
        self._molecule_counts = []
        self._atom_counts = []
        self._site_counts = []
        self._bond_counts = []
        self.add_molecules(molecules)

    def add_molecules(self, molecules):
        api.validate_sequence(molecules, tuple, "molecules",
                              ((lambda p: len(p) == 2, "must have length 2"),
                               (lambda p: isinstance(p[0], Fragment)
                                          and isinstance(p[1], int),
                                "elements must be (fragment, count) pairs")))
        self._molecules.extend(molecules)
        self._templates.extend([f for f, c in molecules])
        self._molecule_counts.extend([c for f, c in molecules])
        self._atom_counts.extend([f.number_of_atoms for f, c in molecules])
        self._site_counts.extend([f.number_of_sites for f, c in molecules])
        self._bond_counts.extend([f.number_of_bonds for f, c in molecules])

    def __eq__(self, other):
        return isinstance(other, Universe) \
               and self.cell_shape == other.cell_shape \
               and self.symmetry_transformations == \
                              other.symmetry_transformations \
               and self.convention == other.convention \
               and all(f1 == f2 and c1 == c2
                       for (f1, c1), (f2, c2)
                            in zip(self.molecules, other.molecules))

    def __hash__(self):
        return hash(self.cell_shape) \
               + len(self.molecules)

    def __str__(self):
        mol_str = ', '.join("(%s('%s'), %d)"
                            % (f.__class__.__name__, f._label, c)
                            for f, c in self._molecules)
        return "Universe('%s', [%s])" % (self.cell_shape, mol_str)

    # Data Model API implementation

    @property
    def cell_shape(self):
        return self._cell_shape

    @cell_shape.setter
    def cell_shape(self, new_cell_shape):
        api.validate_type(new_cell_shape, str, "cell_shape")
        api.validate_value(new_cell_shape,
                           self._cell_parameter_array_shapes.keys(),
                           "cell_shape")
        self._cell_shape = new_cell_shape

    @property
    def symmetry_transformations(self):
        return self._symmetry_transformations

    @symmetry_transformations.setter
    def symmetry_transformations(self, symmetry_transformations):
        if self.cell_shape == "infinite" \
           and len(symmetry_transformations) > 0:
            raise ValueError("Symmetry transformations are allowed "
                             "only in periodic universes")
        api.validate_sequence(symmetry_transformations, tuple,
                              "symmetry_transformations",
                              ((lambda p: len(p) == 2, "must have length 2"),
                               (lambda p: hasattr(p[0], 'shape')
                                          and p[0].shape == (3, 3)
                                          and p[0].dtype == N.float64,
                                "rotation matrix must be float64 "
                                "and have shape (3,3)"),
                               (lambda p: hasattr(p[1], 'shape')
                                          and p[1].shape == (3,)
                                          and p[0].dtype == N.float64,
                                "translation vector must be float64 "
                                "and have shape (3,)"),))
        self._symmetry_transformations = symmetry_transformations

    @property
    def convention(self):
        return self._convention

    @convention.setter
    def convention(self, new_convention):
        api.validate_label(new_convention, "convention")
        self._convention = new_convention

    @property
    def molecules(self):
        return self._molecules

    # More efficient implementations of the number properties

    @property
    def number_of_molecules(self):
        return sum(self._molecule_counts)

    @property
    def number_of_atoms(self):
        return sum(c * na for c, na in zip(self._molecule_counts,
                                           self._atom_counts))

    @property
    def number_of_sites(self):
        return sum(c * ns for c, ns in zip(self._molecule_counts,
                                           self._site_counts))

    @property
    def number_of_bonds(self):
        return sum(c * nb for c, nb in zip(self._molecule_counts,
                                           self._bond_counts))

    @property
    def number_of_template_atoms(self):
        return sum(self._atom_counts)

    @property
    def number_of_template_sites(self):
        return sum(self._site_counts)


class Property(MosaicObject, api.MosaicProperty):

    def __init__(self, universe, name, units,
                 data=None, dtype=None, element_shape=None):
        api.validate_type(universe, Universe, "universe")
        self._universe = universe
        api.validate_label(name, "name")
        self._name = name
        api.validate_units(units, "units")
        self._units = units
        if data is None:
            if dtype is None:
                dtype = N.float64
            if dtype not in self._allowed_dtypes:
                raise ValueError("dtype must be " +
                                 " or ".join(str(t)
                                             for t in self._allowed_dtypes))
            if element_shape is None:
                element_shape = ()
            array_shape = (self._number_of_values(),) + element_shape

            self._data = N.empty(array_shape, dtype)
        else:
            api.validate_array(data, None, self._allowed_dtypes, "data")
            if dtype is not None and data.dtype != dtype:
                raise ValueError("elements of data array do not have "
                                 "the requested type")
            if data.shape[0] != self._number_of_values():
                raise ValueError("data array has incorrect shape")
            if element_shape is not None and data.shape[1:] != element_shape:
                raise ValueError("elements of data array do not have "
                                 "the requested shape")
            self._data = data

    def __str__(self):
        return "%s()" % (self.__class__.__name__)

    def __eq__(self, other):
        return isinstance(other, Property) \
               and self.type == other.type \
               and self.universe == other.universe \
               and self.name == other.name \
               and self.units == other.units \
               and (self.data == other.data).all()

    def __hash__(self):
        return hash(self.type) + hash(self.universe) + hash(self.data.shape)

    # Data Model API implementation
    @property
    def universe(self):
        return self._universe

    @property
    def name(self):
        return self._name

    @property
    def units(self):
        return self._units

    @property
    def data(self):
        return self._data


class AtomProperty(Property):

    def _number_of_values(self):
        return self._universe.number_of_atoms

    # Data Model API implementation
    @property
    def type(self):
        return 'atom'


class SiteProperty(Property):

    def _number_of_values(self):
        return self._universe.number_of_sites

    # Data Model API implementation
    @property
    def type(self):
        return 'site'


class TemplateAtomProperty(Property):

    def _number_of_values(self):
        return self._universe.number_of_template_atoms

    # Data Model API implementation
    @property
    def type(self):
        return 'template_atom'


class TemplateSiteProperty(Property):

    def _number_of_values(self):
        return self._universe.number_of_template_sites

    # Data Model API implementation
    @property
    def type(self):
        return 'template_site'


class Label(MosaicObject, api.MosaicLabel):

    def __init__(self, universe, name, strings=None):
        api.validate_type(universe, Universe, "universe")
        self._universe = universe
        api.validate_label(name, "name")
        self._name = name
        if strings is None:
            self._strings = self._number_of_values() * ['']
        else:
            api.validate_sequence(strings, str, "strings")
            if len(strings) != self._number_of_values():
                raise ValueError("incorrect number of strings")
            for s in strings:
                api.validate_ascii_string(s, "label")
            self._strings = strings

    def __str__(self):
        return "%s()" % (self.__class__.__name__)

    def __eq__(self, other):
        return isinstance(other, Label) \
               and self.type == other.type \
               and self.name == other.name \
               and self.universe == other.universe \
               and self.strings == other.strings

    def __hash__(self):
        return hash(self.type) + hash(self.name) + \
               hash(self.universe) + hash(self.strings)

    # Data Model API implementation
    @property
    def universe(self):
        return self._universe

    @property
    def name(self):
        return self._name

    @property
    def strings(self):
        return self._strings


class AtomLabel(Label):

    def _number_of_values(self):
        return self._universe.number_of_atoms

    # Data Model API implementation
    @property
    def type(self):
        return 'atom'


class SiteLabel(Label):

    def _number_of_values(self):
        return self._universe.number_of_sites

    # Data Model API implementation
    @property
    def type(self):
        return 'site'


class TemplateAtomLabel(Label):

    def _number_of_values(self):
        return self._universe.number_of_template_atoms

    # Data Model API implementation
    @property
    def type(self):
        return 'template_atom'


class TemplateSiteLabel(Label):

    def _number_of_values(self):
        return self._universe.number_of_template_sites

    # Data Model API implementation
    @property
    def type(self):
        return 'template_site'


class Configuration(MosaicObject, api.MosaicConfiguration):

    def __init__(self, universe, positions=None, cell_parameters=None,
                 dtype=None):
        api.validate_type(universe, Universe, "universe")
        cell_param_shape = universe.cell_parameter_array_shape
        nsites = universe.number_of_sites
        if positions is None and cell_parameters is None:
            if dtype is None:
                dtype = N.float64
            if dtype not in self._allowed_dtypes:
                raise ValueError("dtype must be " +
                                 " or ".join(str(t)
                                             for t in self._allowed_dtypes))
            positions = N.empty((nsites, 3), dtype)
            cell_parameters = N.empty(cell_param_shape, dtype)
        else:
            # Allow cell_parameters=None for universes that don't
            # need cell parameters.
            if positions is not None and cell_parameters is None \
               and cell_param_shape == (0,):
                cell_parameters = N.empty(cell_param_shape,
                                          positions.dtype)
            # Require both positions and cell parameters, or neither
            if positions is None or cell_parameters is None:
                raise ValueError("configuration requires both "
                                 "positions and cell parameters")
            # At this point, positions and cell_parameters must be arrays
            # of the required shapes.
            api.validate_array(positions, (nsites, 3),
                               self._allowed_dtypes, "positions")
            api.validate_array(cell_parameters, cell_param_shape,
                               self._allowed_dtypes, "cell_parameters")
            if positions.dtype != cell_parameters.dtype:
                raise ValueError("positions and cell parameters must have"
                                 " the same element type")
            # If an explicit dtype is given, check arrays for conformance
            if dtype is not None:
                if positions.dtype != dtype:
                    raise ValueError("arrays don't have the requested"
                                     " element type " + str(dtype))

        self._universe = universe
        self._positions = positions
        self._cell_parameters = cell_parameters

    def __str__(self):
        return "Configuration()"

    def __eq__(self, other):
        return isinstance(other, Configuration) \
               and self.universe == other.universe \
               and (self.cell_parameters == other.cell_parameters).all() \
               and (self.positions == other.positions).all()

    def __hash__(self):
        return hash(self.universe) \
               + hash(self.positions.shape) + hash(self.cell_parameters.shape)

    # Data Model API implementation
    @property
    def universe(self):
        return self._universe

    @property
    def cell_parameters(self):
        return self._cell_parameters

    @property
    def positions(self):
        return self._positions


class Selection(MosaicObject, api.MosaicSelection):

    def __init__(self, universe, indices):
        api.validate_type(universe, Universe, "universe")
        self._universe = universe
        indices = N.array(indices, uint_for_max_value(max(indices)))
        api.validate_indices(indices, self._max_index(), "indices")
        self._indices = indices

    def __str__(self):
        return "%s()" % (self.__class__.__name__)

    def __eq__(self, other):
        return isinstance(other, Selection) \
               and self.type == other.type \
               and self.universe == other.universe \
               and (self.indices == other.indices).all()

    def __hash__(self):
        return hash(self.type) + hash(self.universe) + hash(len(self.indices))

    # Data Model API implementation
    @property
    def universe(self):
        return self._universe

    @property
    def indices(self):
        return self._indices


class AtomSelection(Selection):

    def _max_index(self):
        return self._universe.number_of_atoms

    # Data Model API implementation
    @property
    def type(self):
        return 'atom'


class SiteSelection(Selection):

    def _max_index(self):
        return self._universe.number_of_sites

    # Data Model API implementation
    @property
    def type(self):
        return 'site'


class TemplateAtomSelection(Selection):

    def _max_index(self):
        return self._universe.number_of_template_atoms

    # Data Model API implementation
    @property
    def type(self):
        return 'template_atom'


class TemplateSiteSelection(Selection):

    def _max_index(self):
        return self._universe.number_of_template_sites

    # Data Model API implementation
    @property
    def type(self):
        return 'template_site'


class Factory(AbstractFactory):

    handler = MethodRegister()

    @handler(api.MosaicUniverse)
    def _make_universe(self, universe):

        def atom(a):
            klass = {"dummy": Dummy,
                     "unknown": Unknown,
                     "element": Element}[a.type]
            return Atom(a.label, klass(a.name), a.number_of_sites)

        def fragment(f):
            fragments = [fragment(sub) for sub in f.fragments]
            atoms = [atom(a) for a in f.atoms]
            if f.is_polymer:
                return Polymer(f.label, f.species,
                               fragments, f.bonds, f.polymer_type)
            else:
                return Fragment(f.label, f.species,
                                fragments, atoms, f.bonds)

        return Universe(universe.cell_shape,
                        [(fragment(f), n) for f, n in universe.molecules],
                        universe.symmetry_transformations,
                        universe.convention)

    @handler(api.MosaicProperty)
    def _make_property(self, property):
        universe = self(property.universe)
        klass = {"atom": AtomProperty,
                 "site": SiteProperty,
                 "template_atom": TemplateAtomProperty,
                 "template_site": TemplateSiteProperty}[property.type]
        return klass(universe, property.name, property.units,
                     property.data)

    @handler(api.MosaicLabel)
    def _make_label(self, label):
        universe = self(label.universe)
        klass = {"atom": AtomLabel,
                 "site": SiteLabel,
                 "template_atom": TemplateAtomLabel,
                 "template_site": TemplateSiteLabel}[label.type]
        return klass(universe, label.name, label.strings)

    @handler(api.MosaicConfiguration)
    def _make_configuration(self, configuration):
        universe = self(configuration.universe)
        return Configuration(universe,
                             configuration.positions,
                             configuration.cell_parameters)

    @handler(api.MosaicSelection)
    def _make_selection(self, selection):
        universe = self(selection.universe)
        klass = {"atom": AtomSelection,
                 "site": SiteSelection,
                 "template_atom": TemplateAtomSelection,
                 "template_site": TemplateSiteSelection}[selection.type]
        return klass(universe, selection.indices)




# -*- coding: utf-8 -*-

# Tests for Mosaic utility functions

#-----------------------------------------------------------------------------
#       Copyright (C) 2013 The Mosaic Development Team
#
#  Distributed under the terms of the BSD License.  The full license is in
#  the file LICENSE.txt, distributed as part of this software.
#-----------------------------------------------------------------------------

import unittest

import numpy as N

from mosaic.api import validate_units

def valid_units(units):
    try:
        validate_units(units, "")
        return True
    except (ValueError, TypeError):
        return False

class UnitStringTest(unittest.TestCase):

    def test_units(self):
        self.assertTrue(valid_units("nm"))
        self.assertTrue(valid_units("nm ps-1"))
        self.assertTrue(valid_units("g mol-1"))
        self.assertTrue(valid_units("60 s"))
        self.assertTrue(valid_units("1000 m"))
        self.assertTrue(valid_units("0.001 m"))
        # Not a string
        self.assertFalse(valid_units(42))
        # Invalid symbols
        self.assertFalse(valid_units("foo"))
        self.assertFalse(valid_units("m foo-2"))
        # Two numerical factors
        self.assertFalse(valid_units("10 10 m"))
        # Numerical factor not in initial position
        self.assertFalse(valid_units("m 10 s-1"))
        # Non-integer exponent
        self.assertFalse(valid_units("m2.5"))
        # Numerical factor neither integer nor decimal fraction
        self.assertFalse(valid_units("1.2e5 s"))

def suite():
    loader = unittest.TestLoader()
    s = unittest.TestSuite()
    s.addTest(loader.loadTestsFromTestCase(UnitStringTest))
    return s

if __name__ == '__main__':
    unittest.main()





# -*- coding: utf-8 -*-
"""HDF5 I/O

.. moduleauthor:: Konrad Hinsen

Any model implementing the Mosaic API can be written to a HDF5 file,
but will be implicitly converted to a :mod:`mosaic.array_model`
representation.  Data items read from a HDF5 file are always
represented by an :mod:`mosaic.array_model`.

The HDF5 storage layout is designed for clarity rather than
simplicity, the files should be as self-documenting as possible.

A common pattern for generating an HDF5 file is:

::

    with HDF5Store('molecule.h5', 'w') as writer:
       writer.store("universe", universe)
       writer.store("configuration1", configuration1)
       writer.store("configuration2", configuration2)

A common pattern for reading from an HDF5 file is:

::

    items = {}
    for id, data in HDF5Store('molecule.h5'):
       items[id] = data

"""

#-----------------------------------------------------------------------------
#       Copyright (C) 2013 The Mosaic Development Team
#
#  Distributed under the terms of the BSD License.  The full license is in
#  the file LICENSE.txt, distributed as part of this software.
#-----------------------------------------------------------------------------

import copy
import operator
import weakref

import h5py
import h5py.h5t
import h5py.h5d
import h5py.h5s
import h5py.version
import numpy as N

import mosaic.array_model
import mosaic.api as api
from mosaic.utility import MethodRegister
from mosaic.utility import ascii
from mosaic.utility import isstring
from mosaic.utility import py_str

import sys
del sys


if h5py.version.hdf5_version_tuple < (1, 8, 7):
    raise ValueError("HDF5 version >= 1.8.7 required")


def create_dataset(group, path, data):
    if isstring(data):
        ds = group.require_dataset(path, shape=(),
                                   dtype=h5py.special_dtype(vlen=bytes))
        ds[...] = data
    else:
        ds = group.require_dataset(path, shape=data.shape, dtype=data.dtype)
        if 0 not in data.shape:
            ds[...] = data
    return ds


class HDF5Store(object):

    """I/O handler for HDF5 files

    This class handles the translation of references between Mosaic
    data items in memory and HDF5 object references in a file. When
    writing data, references are allowed only to data items that have
    been written earlier using the same HDF5Store instance.

    An HDF5Store can be used like a file, in which case it must be
    closed after the last ``store`` or ``retrieve`` operation, or as a
    context manager in a with-statement. For reading, it also behaves
    like an iterator yielding ``(name, data_item)`` pairs.
    """

    def __init__(self, hdf5_group_or_filename, file_mode=None):
        """
        :param hdf5_group_or_filename: a group in an open HDF5 file,
                                       or a string interpreted as a file name
        :type hdf5_group_or_filename: str or ``h5py.Group``
        :param file_mode: ``'r'`` or ``'w'``, used only with a filename argument
        :type file_mode:  str
        """
        if isstring(hdf5_group_or_filename):
            if file_mode is None:
                file_mode = 'r'
            self.root = h5py.File(hdf5_group_or_filename, file_mode)
            self._close = True
        else:
            api.validate_value(file_mode, [None], "file_mode")
            self.root = hdf5_group_or_filename
            self._close = False
        self._path_map = weakref.WeakKeyDictionary()
        self._data_map = weakref.WeakValueDictionary()
        self._factory = mosaic.array_model.Factory()

    def close(self):
        if self._close:
            self.root.file.close()
        self.root = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()
        return False

    def flush(self):
        self.root.file.flush()

    storage_handler = MethodRegister()

    def store(self, path, data):
        """
        :param path: a HDF5 path, relative to the group used by the HDF5Store
        :type path:  str
        :param data: a Mosaic data item
        :type data:  :class:`mosaic.api.MosaicDataItem`
        """
        api.validate_type(path, str, "path")
        if path[0] == "/":
            raise ValueError("absolute paths not allowed")
        api.validate_type(data, api.MosaicDataItem, "data")
        handler = self.storage_handler[type(data)]
        if handler is None:
            raise TypeError("Storage of %s not yet implemented"
                            % str(type(data)))
        handler(self, path, data)
        self._register_data_item(path, data)

    def retrieve(self, path_or_node):
        """
        :param path_or_node: a HDF5 path, relative to the group used by
                             the HDF5Store, or an HDF5 node
        :type path_or_node:  str or h5py.Node
        :returns: a Mosaic data item
        :rtype:  :class:`mosaic.api.MosaicDataItem`
        """
        if isinstance(path_or_node, h5py.h5r.Reference):
            path_or_node = self.root[path_or_node]
        if isinstance(path_or_node, str):
            path = path_or_node
        else: # Group, Dataset, or classes with the same interface
            path = path_or_node.name
        if path[0] == '/':
            if path.startswith(self.root.name):
                path = path[len(self.root.name):]
                if path[0] == '/':
                    path = path[1:]
            else:
                raise ValueError("path name not inside root group")
        data = self._get_data(path)
        if data is not None:
            return data
        node = self.root[path]
        label = py_str(self._read_stamp(node))
        if label not in self._node_types:
            raise ValueError("Undefined node type %s" % label)
        handler = self.storage_handler[label]
        if handler is None:
            raise ValueError("Retrieval of node type %s not yet implemented"
                             % label)
        data = handler(self, node)
        self._register_data_item(path, data)
        return data

    _node_types = ['universe', 'configuration', 'property', 'label',
                   'selection']

    def __iter__(self):
        # First pass: universes
        for name in self.root:
            node = self.root[name]
            if node.attrs.get('DATA_MODEL', None) == ascii('MOSAIC') \
               and node.attrs['MOSAIC_DATA_TYPE'] == ascii('universe'):
                yield name, self.retrieve(node)
        # Second pass: anything else
        for name in self.root:
            node = self.root[name]
            if node.attrs.get('DATA_MODEL', None) == ascii('MOSAIC') \
               and node.attrs['MOSAIC_DATA_TYPE'] != ascii('universe'):
                yield name, self.retrieve(node)

    def recursive_iterator(self, group=None):
        if group is None:
            group = self.root
        # First pass: universes
        for name in group:
            node = group[name]
            if node.attrs.get('DATA_MODEL', None) == ascii('MOSAIC') \
               and node.attrs['MOSAIC_DATA_TYPE'] == ascii('universe'):
                yield name, self.retrieve(node)
            if isinstance(node, h5py.Group):
                for gname, gnode in self.recursive_iterator(node):
                    yield name+'/'+gname, gnode
        # Second pass: anything else
        for name in group:
            node = group[name]
            if node.attrs.get('DATA_MODEL', None) == ascii('MOSAIC') \
               and node.attrs['MOSAIC_DATA_TYPE'] != ascii('universe'):
                yield name, self.retrieve(node)
            if isinstance(node, h5py.Group):
                for gname, gnode in self.recursive_iterator(node):
                    yield name+'/'+gname, gnode

    def _stamp(self, node, label):
        node.attrs['DATA_MODEL'] = ascii('MOSAIC')
        node.attrs['DATA_MODEL_MAJOR_VERSION'] = api.MOSAIC_VERSION[0]
        node.attrs['DATA_MODEL_MINOR_VERSION'] = api.MOSAIC_VERSION[1]
        node.attrs['MOSAIC_DATA_TYPE'] = ascii(label)

    def _read_stamp(self, node):
        if node.attrs['DATA_MODEL'] != ascii('MOSAIC'):
            raise ValueError("Node %s not an Mosaic data item" % str(node))
        version = (node.attrs['DATA_MODEL_MAJOR_VERSION'],
                   node.attrs['DATA_MODEL_MINOR_VERSION'])
        if version[0] > api.MOSAIC_VERSION[0] \
           or version[1] > api.MOSAIC_VERSION[1]:
            raise ValueError("HDF5 data is for version %d.%d, "
                             "software is version %d.%d"
                             % (version + api.MOSAIC_VERSION))
        return node.attrs['MOSAIC_DATA_TYPE']

    def _register_data_item(self, path, data_item):
        self._data_map[path] = data_item
        self._path_map[data_item] = path

    def _get_path(self, data_item):
        return self._path_map.get(data_item, None)

    def _get_data(self, path):
        return self._data_map.get(path, None)

    #
    # Store data in HDF5 file
    #
    @storage_handler(api.MosaicUniverse)
    def _store_universe(self, path, universe):
        group = self.root.require_group(path)
        try:
            # ActivePapers support
            group.mark_as_data_item()
        except AttributeError:
            pass
        self._stamp(group, 'universe')

        # Convert the universe to an array_model
        universe = self._factory(universe)

        # Create HDF5 datasets
        create_dataset(group, 'cell_shape',
                       data=universe.cell_shape)
        create_dataset(group, 'symmetry_transformations',
                       data=N.array(universe._symmetry_transformations))
        create_dataset(group, 'convention',
                       data=universe.convention)
        create_dataset(group, 'fragments', data=universe._fragment_array)
        create_dataset(group, 'atoms', data=universe._atom_array)
        create_dataset(group, 'bonds', data=universe._bond_array)
        create_dataset(group, 'molecules', data=universe._molecule_array)
        if len(universe._polymer_array) > 0:
            create_dataset(group, 'polymers', data=universe._polymer_array)
        symbols = universe._symbols
        symbol_table = group.require_dataset('symbols',
                                shape=(len(symbols),),
                                dtype=h5py.special_dtype(vlen=bytes))
        for i, s in enumerate(symbols):
            symbol_table[i] = ascii(s)

    @storage_handler(api.MosaicConfiguration)
    def _store_configuration(self, path, configuration):
        universe = configuration.universe
        universe_path = self._get_path(universe)
        if universe_path is None:
            raise IOError("universe must be stored first")
        group = self.root.require_group(path)
        try:
            # ActivePapers support
            group.mark_as_data_item()
        except AttributeError:
            pass
        self._stamp(group, 'configuration')
        group.attrs['universe'] = self.root[universe_path].ref
        ncellparams = N.multiply.reduce(universe.cell_parameter_array_shape)
        if ncellparams > 0:
            create_dataset(group, 'cell_parameters',
                           data=configuration.cell_parameters)
        arr = N.ascontiguousarray(configuration.positions)
        positions = group.require_dataset('positions',
                                          shape=(len(arr),),
                                          dtype='(3,)f%d' % arr.dtype.itemsize)
        # There doesn't seem to be any way to write this array
        # using high-level operations, so we use the low-level access.
        mtype = h5py.h5t.py_create(positions.id.dtype)
        mspace = h5py.h5s.create_simple(positions.shape)
        fspace = positions.id.get_space()
        positions.id.write(mspace, fspace, arr, mtype)

    @storage_handler(api.MosaicProperty)
    def _store_property(self, path, property):
        universe = property.universe
        universe_path = self._get_path(universe)
        if universe_path is None:
            raise IOError("universe must be stored first")
        element_shape = property.data.shape[1:]
        if element_shape:
            dtype = N.dtype((property.data.dtype, element_shape))
        else:
            dtype = property.data.dtype
        arr = N.ascontiguousarray(property.data)
        ds = self.root.require_dataset(path,
                                       shape=(len(arr),),
                                       dtype=dtype)
        self._stamp(ds, 'property')
        ds.attrs['universe'] = self.root[universe_path].ref
        ds.attrs['name'] = ascii(property.name)
        ds.attrs['units'] = ascii(property.units)
        ds.attrs['property_type'] = ascii(property.type)
        # There doesn't seem to be any way to write this array
        # using high-level operations, so we use the low-level access.
        mtype = h5py.h5t.py_create(ds.id.dtype)
        mspace = h5py.h5s.create_simple(ds.shape)
        fspace = ds.id.get_space()
        ds.id.write(mspace, fspace, arr, mtype)

    @storage_handler(api.MosaicLabel)
    def _store_label(self, path, label):
        universe = label.universe
        label = self._factory(label)
        universe_path = self._get_path(universe)
        if universe_path is None:
            raise IOError("universe must be stored first")
        ds = create_dataset(self.root, path, data=label._string_array)
        self._stamp(ds, 'label')
        ds.attrs['universe'] = self.root[universe_path].ref
        ds.attrs['name'] = ascii(label.name)
        ds.attrs['label_type'] = ascii(label.type)

    @storage_handler(api.MosaicSelection)
    def _store_selection(self, path, selection):
        universe = selection.universe
        universe_path = self._get_path(universe)
        if universe_path is None:
            raise IOError("universe must be stored first")
        ds = create_dataset(self.root, path,
                            N.ascontiguousarray(selection.indices))
        ds.attrs['universe'] = self.root[universe_path].ref
        ds.attrs['selection_type'] = ascii(selection.type)
        self._stamp(ds, 'selection')

    #
    # Retrieve data from HDF5 file
    #
    @storage_handler('universe')
    def _retrieve_universe(self, node):
        # Work around a restriction/bug in h5py that prevents reading
        # arrays of length zero.
        if node['symmetry_transformations'].shape[0] == 0:
            st = N.zeros((0,), dtype=node['symmetry_transformations'].dtype)
        else:
            st = node['symmetry_transformations'][...]
        return mosaic.array_model.Universe.from_arrays(
                        node['atoms'][...],
                        node['bonds'][...],
                        node['fragments'][...],
                        node['polymers'][...]
                            if 'polymers' in node else None,
                        node['molecules'][...],
                        tuple(py_str(s) for s in node['symbols'][...]),
                        st,
                        py_str(node['cell_shape'][()]),
                        py_str(node['convention'][()]))

    @storage_handler('configuration')
    def _retrieve_configuration(self, node):
        universe = self.retrieve(node.attrs['universe'])
        positions = node['positions'][...]
        if 'cell_parameters' in node:
            ds = node['cell_parameters']
            cell_parameters = N.empty(ds.shape, ds.dtype)
            cell_parameters[...] = ds[...]
        else:
            assert N.product(universe.cell_parameter_array_shape) == 0
            cell_parameters = N.empty(universe.cell_parameter_array_shape,
                                      positions.dtype)
        return mosaic.array_model.Configuration(universe, positions,
                                                cell_parameters)

    @storage_handler('property')
    def _retrieve_property(self, node):
        universe = self.retrieve(node.attrs['universe'])
        data = node[...]
        name = py_str(node.attrs['name'])
        units = py_str(node.attrs['units'])
        property_type = py_str(node.attrs['property_type'])
        return mosaic.array_model.Property(universe, name, units,
                                           data, property_type)

    @storage_handler('label')
    def _retrieve_label(self, node):
        universe = self.retrieve(node.attrs['universe'])
        data = node[...]
        name = py_str(node.attrs['name'])
        label_type = py_str(node.attrs['label_type'])
        return mosaic.array_model.Label.from_arrays(universe, name,
                                                    data, label_type)

    @storage_handler('selection')
    def _retrieve_selection(self, node):
        universe = self.retrieve(node.attrs['universe'])
        indices = node[...]
        selection_type = py_str(node.attrs['selection_type'])
        return mosaic.array_model.Selection(universe, indices,
                                            selection_type)



# -*- coding: utf-8 -*-

# Tests for module mosaic.immutable_model

#-----------------------------------------------------------------------------
#       Copyright (C) 2013 The Mosaic Development Team
#
#  Distributed under the terms of the BSD License.  The full license is in
#  the file LICENSE.txt, distributed as part of this software.
#-----------------------------------------------------------------------------

import unittest

import numpy as N
import immutable.np as IN

import mosaic.immutable_model as M
from mosaic.api import is_valid

def make_water_fragment(nsites=1):
    return M.fragment("water", (),
                      (("H1", M.atom(M.element("H"), nsites)),
                       ("H2", M.atom(M.element("H"), nsites)),
                       ("O",  M.atom(M.element("O"), nsites))),
                      (("H1", "O", "single"), ("H2", "O", "single")))


class AtomDescriptorTest(unittest.TestCase):

    def test_singleton(self):
        self.assertTrue(M.dummy() is M.dummy())
        self.assertTrue(M.dummy('a') is M.dummy('a'))
        self.assertTrue(M.dummy('a') is not M.dummy('b'))
        self.assertTrue(M.dummy('a') is not M.unknown('a'))
        self.assertTrue(M.dummy('C') is not M.element('C'))
        self.assertTrue(M.element('C') is M.element('C'))

    def test_name(self):
        for name in ['a', 'b', 'c']:
            self.assertEqual(M.unknown(name).name, name)

    def test_type(self):
        self.assertEqual(M.dummy().type, "dummy")
        self.assertEqual(M.unknown().type, "")
        self.assertEqual(M.element('O').type, "element")
        self.assertEqual(M.cgparticle('ala').type, "cgparticle")

class WaterTest(unittest.TestCase):

    def setUp(self):
        self.mol = make_water_fragment()

    def test_basics(self):
        self.assertEqual(self.mol.number_of_atoms, 3)
        self.assertEqual(self.mol.number_of_sites, 3)
        self.assertEqual(self.mol.number_of_bonds, 2)
        self.assertEqual(self.mol.species, "water")

    def test_equality(self):
        same_mol = make_water_fragment()
        changed_bond_order = M.fragment("water", (),
                                        (("H1", M.atom(M.element("H"))),
                                         ("H2", M.atom(M.element("H"))),
                                         ("O",  M.atom(M.element("O")))),
                                        (("O", "H2", "single"),
                                         ("O", "H1", "single")))
        changed_atom_order = M.fragment("water", (),
                                        (("O",  M.atom(M.element("O"))),
                                         ("H1", M.atom(M.element("H"))),
                                         ("H2", M.atom(M.element("H")))),
                                        (("O", "H1", "single"),
                                         ("O", "H2", "single")))
        self.assertEqual(self.mol, self.mol)
        self.assertEqual(self.mol, same_mol)
        self.assertEqual(self.mol, changed_bond_order)
        self.assertNotEqual(self.mol, changed_atom_order)
        
class PeptideTest(unittest.TestCase):

    def _make_molecule(self):
        C = M.element('C')
        H = M.element('H')
        N = M.element('N')
        O = M.element('O')
        peptide_group = M.fragment('peptide',
                                   (),
                                   (('CA', M.atom(C)),
                                    ('HA', M.atom(H)),
                                    ('H', M.atom(H)),
                                    ('N', M.atom(N)),
                                    ('C', M.atom(C)),
                                    ('O', M.atom(O))),
                                   (('N', 'H', "single"),
                                    ('N', 'CA', "single"),
                                    ('CA', 'HA', "single"),
                                    ('CA', 'C', "single"),
                                    ('C', 'O', "double")))
        ala_sidechain = M.fragment('ala_sidechain',
                                   (),
                                   (('CB', M.atom(C)),
                                    ('HB1', M.atom(H)),
                                    ('HB2', M.atom(H)),
                                    ('HB3', M.atom(H))),
                                   (('CB', 'HB1', "single"),
                                    ('CB', 'HB2', "single"),
                                    ('CB', 'HB3', "single"),))
        ala = M.fragment('alanine',
                         (('peptide', peptide_group),
                          ('sidechain', ala_sidechain)),
                         (),
                         (('peptide.CA', 'sidechain.CB', "single"),))
        return M.polymer('alanine_dipeptide',
                         (('ALA1', ala),
                          ('ALA2', ala)),
                         (('ALA1.peptide.C', 'ALA2.peptide.N', "single"),),
                         'polypeptide')

    def test_basic(self):
        mol = self._make_molecule()
        self.assertEqual(mol.number_of_atoms, 20)
        self.assertEqual(mol.number_of_sites, 20)
        self.assertEqual(mol.number_of_bonds, 19)
        self.assertEqual(mol.polymer_type, "polypeptide")

    def test_equality(self):
        self.assertEqual(self._make_molecule(),
                         self._make_molecule())

    def test_iterators(self):
        mol = self._make_molecule()
        mol_ref = M.FragmentRef('x', mol)
        atoms = tuple(mol_ref.recursive_atom_iterator())
        self.assertEqual(len(atoms), mol.number_of_atoms)
        bonds = tuple(mol_ref.recursive_bond_iterator())
        self.assertEqual(len(bonds), mol.number_of_bonds)
        for a1, a2, order in bonds:
            for a in a1, a2:
                node = mol
                for p in a.split('.'):
                    node = node[p]
                self.assertTrue(isinstance(node, M.Atom))
        paths = tuple(mol_ref.recursive_atom_path_iterator())
        self.assertEqual(len(paths), mol.number_of_atoms)
        for ap in paths:
            node = mol
            for p in ap.split('.'):
                node = node[p]
            self.assertTrue(isinstance(node, M.Atom))

class ErrorCheckingTest(unittest.TestCase):

    def test_atom_descriptor(self):
        self.assertRaises(TypeError, lambda: M.dummy(42))
        self.assertRaises(ValueError, lambda: M.element(42))
        self.assertRaises(ValueError, lambda: M.element("X"))
        
    def test_atom(self):
        carbon = M.element("C")
        self.assertRaises(TypeError, lambda: M.atom('C', 1))
        self.assertRaises(ValueError, lambda: M.atom(carbon, 0))
        
    def test_fragment(self):
        carbon = M.atom(M.element("C"))
        # Illegal fragments
        self.assertRaises(TypeError,
                          lambda: M.fragment('m', None, (("C", carbon),), ()))
        self.assertRaises(TypeError,
                          lambda: M.fragment('m', [1, 2], (("C", carbon),), ()))
        self.assertRaises(TypeError,
                          lambda: M.fragment('m', (("C", carbon),), (), ()))
        # Illegal atoms
        self.assertRaises(TypeError,
                          lambda: M.fragment('m', (), None, ()))
        self.assertRaises(TypeError,
                          lambda: M.fragment('m', (), [1, 2], ()))
        self.assertRaises(TypeError,
                          lambda: M.fragment('m', (), (carbon,), ()))
        self.assertRaises(ValueError,
                          lambda: M.fragment('m', (),
                                             (("C", carbon),
                                              ("C", carbon)),
                                             ()))
        # Illegal bond lists
        self.assertRaises(TypeError,
                          lambda: M.fragment('m', (), (("C", carbon),), None))
        self.assertRaises(TypeError,
                          lambda: M.fragment('m', (), (("C", carbon),),
                                             [1, 2, 3]))
        self.assertRaises(TypeError,
                          lambda: M.fragment('m', (), (("C", carbon),),
                                             (('X', 'X'))))
        self.assertRaises(TypeError,
                          lambda: M.fragment('m', (), (("C", carbon),),
                                             (['X', 'X', 'single'])))
        
    def test_bonds(self):
        carbon = M.atom(M.element("C"))
        # Bond specified by only one atom
        self.assertRaises(ValueError,
                          lambda: M.fragment('m', (),
                                             (('C1', carbon), ('C2', carbon)),
                                             (('C1', ),)))
        # Bond specified by two atoms but no bond order
        self.assertRaises(ValueError,
                          lambda: M.fragment('m', (),
                                             (('C1', carbon), ('C2', carbon)),
                                             (('C1', 'C2'),)))
        # Bond specified by two identical atoms
        self.assertRaises(ValueError,
                          lambda: M.fragment('m', (),
                                             (('C1', carbon), ('C2', carbon)),
                                             (('C1', 'C1', ''),)))
        # Bond specified by an atom name that is undefined
        self.assertRaises(ValueError,
                          lambda: M.fragment('m', (),
                                             (('C1', carbon), ('C2', carbon)),
                                             (('C1', 'C3', ''),)))
        # Bond specified at the wrong fragment level
        f = M.fragment('x', (), (('C1', carbon), ('C2', carbon)), ())
        self.assertRaises(ValueError,
                          lambda: M.fragment('m', (('x', f),),
                                             (('C3', carbon),),
                                             (('x.C1', 'x.C2', ''),)))

    def test_universe(self):
        mol = M.fragment("water", (),
                         (("H1", M.atom(M.element("H"), 8)),
                          ("H2", M.atom(M.element("H"), 8)),
                          ("O",  M.atom(M.element("O"), 2))),
                         (("H1", "O", "single"), ("H2", "O", "single")))
        self.assertRaises(TypeError,
                          lambda: M.universe(0, [(mol, 'water', 10)]))
        self.assertRaises(ValueError,
                          lambda: M.universe('strange', [(mol, 'water', 10)]))
        self.assertRaises(ValueError,
                          lambda: M.universe('strange', [(mol, 10)]))
        self.assertRaises(TypeError,
                          lambda: M.universe('infinite', mol))
        self.assertRaises(ValueError,
                          lambda: M.universe('infinite', [("water", 10)]))
        self.assertRaises(TypeError,
                          lambda: M.universe('infinite', [mol]))
        self.assertRaises(ValueError,
                          lambda: M.universe('infinite', [(10, mol)]))
        self.assertRaises(ValueError,
                          lambda: M.universe('infinite', [(mol, 'water', 10)],
                                             [(IN.zeros((3,3), N.float64),
                                               IN.zeros((3,), N.float64))]))

    def test_configuration(self):
        mol = make_water_fragment()
        universe = M.universe('cube', [(mol, 'water', 10)])
        # Missing data
        self.assertRaises(TypeError,
                          lambda: M.Configuration(universe))
        # Positions but no cell parameters
        self.assertRaises(TypeError,
                          lambda: M.Configuration(universe,
                                                  IN.zeros((30, 3), N.float32)))
        # Positions and cell parameters of different dtype
        self.assertRaises(ValueError,
                          lambda: M.Configuration(universe,
                                                  IN.zeros((30, 3), N.float32),
                                                  N.float64(10.)))
        # Positions not an array
        self.assertRaises(TypeError,
                          lambda: M.Configuration(universe,
                                                  list(IN.zeros((30, 3),
                                                               N.float32)),
                                                  N.float32(10.)))
        # Positions of wrong shape
        self.assertRaises(ValueError,
                          lambda: M.Configuration(universe,
                                                  IN.zeros((25, 3), N.float32),
                                                  N.float32(10.)))
        # Cell parameters of wrong shape
        self.assertRaises(ValueError,
                          lambda: M.Configuration(universe,
                                                  IN.zeros((30, 3), N.float32),
                                                  IN.zeros((3,), N.float32)))
        
    def test_selection(self):
        mol = make_water_fragment(nsites=2)
        universe = M.universe('cube', [(mol, 'water', 5)])
        # Index occurs twice
        self.assertRaises(ValueError,
                          lambda: M.AtomSelection(universe,
                                                  IN.zeros((2,), N.uint16)))
        # Atom index too large
        self.assertRaises(ValueError,
                          lambda: M.AtomSelection(universe,
                                                  IN.array([20], N.uint16)))
        # Template atom index too large
        self.assertRaises(ValueError,
                          lambda: M.TemplateAtomSelection(universe,
                                                          IN.array([3], N.uint8)))
        # Site index too large
        self.assertRaises(ValueError,
                          lambda: M.SiteSelection(universe,
                                                  IN.array([40], N.uint16)))
        # Template site index too large
        self.assertRaises(ValueError,
                          lambda: M.TemplateSiteSelection(universe,
                                                          IN.array([8], N.uint8)))

class UniverseTest(unittest.TestCase):

    def setUp(self):
        mol = make_water_fragment(2)
        self.universe = M.universe('infinite', [(mol, 'water', 10)],
                                   convention='my_own')

    def test_basics(self):
        self.assertTrue(is_valid(self.universe))
        self.assertEqual(self.universe.number_of_molecules, 10)
        self.assertEqual(self.universe.number_of_atoms, 30)
        self.assertEqual(self.universe.number_of_sites, 60)
        self.assertEqual(self.universe.number_of_bonds, 20)
        self.assertEqual(self.universe.cell_shape, "infinite")
        self.assertEqual(self.universe.convention, "my_own")

    def test_properties(self):
        masses = M.TemplateAtomProperty(self.universe,
                                        "masses", "amu",
                                        IN.array([1., 1., 16.], N.float32))
        self.assertTrue(is_valid(masses))
        self.assertEqual(masses.type, 'template_atom')
        self.assertTrue(masses.universe == self.universe)
        self.assertEqual(masses.element_shape, ())
        self.assertEqual(masses.data.shape, (3,))
        bead_masses = M.TemplateSiteProperty(self.universe,
                                             "mass", "amu",
                                             IN.array([1., 1.,
                                                       1., 1.,
                                                       8., 8.], N.float32))
        self.assertTrue(is_valid(bead_masses))
        self.assertEqual(bead_masses.type, 'template_site')
        self.assertTrue(bead_masses.universe is self.universe)
        self.assertEqual(bead_masses.element_shape, ())
        self.assertEqual(bead_masses.data.shape, (6,))
        velocities = M.SiteProperty(self.universe,
                                    "velocity", "nm ps-1",
                                    IN.zeros((60, 3), dtype=N.float64))
        self.assertTrue(is_valid(velocities))
        self.assertEqual(velocities.type, 'site')
        self.assertTrue(velocities.universe is self.universe)
        self.assertEqual(velocities.data.shape, (60, 3))
        self.assertEqual(velocities.element_shape, (3,))
        foo = M.AtomProperty(self.universe,
                             "foo", "",
                             IN.zeros((30, 2, 2), dtype=N.int16))
        self.assertTrue(is_valid(foo))
        self.assertEqual(foo.type, 'atom')
        self.assertTrue(foo.universe is self.universe)
        self.assertEqual(foo.data.shape, (30, 2, 2))
        self.assertEqual(foo.element_shape, (2, 2))

    def test_labels(self):
        labels = tuple(a.name
                       for f, n in self.universe.molecules
                       for a in f.recursive_atom_iterator())
        el = M.TemplateAtomLabel(self.universe, "element", labels)
        self.assertTrue(is_valid(el))
        self.assertEqual(el.name, "element")
        self.assertTrue(el.universe == self.universe)
        self.assertTrue(len(el.strings)
                        == self.universe.number_of_template_atoms)
        for s1, s2 in zip(labels, el.strings):
            self.assertEqual(s1, s2)

        labels = tuple(a.name
                       for f, n in self.universe.molecules
                       for _ in range(n)
                       for a in f.recursive_atom_iterator())
        el = M.AtomLabel(self.universe, "element", labels)
        self.assertTrue(is_valid(el))
        self.assertEqual(el.name, "element")
        self.assertTrue(el.universe == self.universe)
        self.assertTrue(len(el.strings)
                        == self.universe.number_of_atoms)
        for s1, s2 in zip(labels, el.strings):
            self.assertEqual(s1, s2)

        labels = tuple(a.name
                       for f, n in self.universe.molecules
                       for a in f.recursive_atom_iterator()
                       for _ in range(a.number_of_sites))
        el = M.TemplateSiteLabel(self.universe, "element", labels)
        self.assertTrue(is_valid(el))
        self.assertEqual(el.name, "element")
        self.assertTrue(el.universe == self.universe)
        self.assertTrue(len(el.strings)
                        == self.universe.number_of_template_sites)
        for s1, s2 in zip(labels, el.strings):
            self.assertEqual(s1, s2)

        labels = tuple(a.name
                       for f, n in self.universe.molecules
                       for _ in range(n)
                       for a in f.recursive_atom_iterator()
                       for __ in range(a.number_of_sites))
        el = M.SiteLabel(self.universe, "element", labels)
        self.assertTrue(is_valid(el))
        self.assertEqual(el.name, "element")
        self.assertTrue(el.universe == self.universe)
        self.assertTrue(len(el.strings)
                        == self.universe.number_of_sites)
        for s1, s2 in zip(labels, el.strings):
            self.assertEqual(s1, s2)

    def test_bonds(self):
        bonds = self.universe.bond_index_array()
        self.assertEqual(len(bonds), self.universe.number_of_bonds)
        self.assertTrue((bonds >= 0).all())
        self.assertTrue((bonds < self.universe.number_of_atoms).all())
        for i in range(10):
            self.assertEqual(bonds[2*i, 0], 3*i)
            self.assertEqual(bonds[2*i, 1], 3*i+2)
            self.assertEqual(bonds[2*i+1, 0], 3*i+1)
            self.assertEqual(bonds[2*i+1, 1], 3*i+2)

    def test_index_mappings(self):
        mol = self.universe.molecules[0][0]
        s2a = mol.site_to_atom_index_mapping()
        self.assertTrue((s2a == N.array([0, 0, 1, 1, 2, 2])).all())

        s2a = self.universe.site_to_atom_index_mapping()
        s2a_ref = N.repeat(N.arange(30), 2)
        self.assertTrue((s2a == s2a_ref).all())

        st2at = self.universe.template_site_to_template_atom_index_mapping()
        st2at_ref = N.array([0, 0, 1, 1, 2, 2])
        self.assertTrue((st2at == st2at_ref).all())

        s2t = self.universe.site_to_template_index_mapping()
        s2t_ref = N.resize(N.arange(mol.number_of_sites),
                           (self.universe.number_of_sites,))
        self.assertTrue((s2t == s2t_ref).all())

        a2t = self.universe.atom_to_template_index_mapping()
        a2t_ref = N.resize(N.arange(mol.number_of_atoms),
                           (self.universe.number_of_atoms,))
        self.assertTrue((a2t == a2t_ref).all())

        a2s = self.universe.atom_to_site_index_mapping()
        a2s_ref = 2*N.arange(self.universe.number_of_atoms)
        self.assertTrue((a2s == a2s_ref).all())

    def test_selections(self):
        s = M.AtomSelection(self.universe, IN.array([0, 2], N.uint8))
        self.assertEqual(s.number_of_atoms, 2)
        self.assertEqual(s.number_of_sites, 4)
        s = M.TemplateAtomSelection(self.universe, IN.array([1], N.uint8))
        self.assertEqual(s.number_of_atoms, 10)
        self.assertEqual(s.number_of_sites, 20)
        s = M.SiteSelection(self.universe, [2])
        self.assertEqual(s.number_of_sites, 1)
        s = M.TemplateSiteSelection(self.universe, [0])
        self.assertEqual(s.number_of_sites, 10)

class PBCTest(unittest.TestCase):

    def setUp(self):
        self.infinite = M.universe('infinite', ())
        self.cube = M.universe('cube', ())
        self.cuboid = M.universe('cuboid', ())
        self.parallelepiped = M.universe('parallelepiped', ())

    def test_lattice_vectors(self):
        conf = M.Configuration(self.infinite,
                               IN.zeros((0, 3), N.float32),
                               None)
        self.assertEqual(conf.lattice_vectors(), ())
        self.assertEqual(conf.cell_volume(), None)
        conf = M.Configuration(self.cube,
                               IN.zeros((0, 3), N.float32),
                               IN.array(1., N.float32))
        lv = conf.lattice_vectors()
        self.assertTrue((lv[0] == N.array([1., 0., 0.], N.float32)).all())
        self.assertTrue((lv[1] == N.array([0., 1., 0.], N.float32)).all())
        self.assertTrue((lv[2] == N.array([0., 0., 1.], N.float32)).all())
        self.assertEqual(conf.cell_volume(), 1.)
        conf = M.Configuration(self.cuboid,
                               IN.zeros((0, 3), N.float32),
                               IN.array([1., 2., 4.], N.float32))
        lv = conf.lattice_vectors()
        self.assertTrue((lv[0] == N.array([1., 0., 0.], N.float32)).all())
        self.assertTrue((lv[1] == N.array([0., 2., 0.], N.float32)).all())
        self.assertTrue((lv[2] == N.array([0., 0., 4.], N.float32)).all())
        self.assertEqual(conf.cell_volume(), 8.)
        conf = M.Configuration(self.parallelepiped,
                               IN.zeros((0, 3), N.float32),
                               IN.array([[1., 2., 4.],
                                         [8., 4., 2.],
                                         [16., 4., 8.]], N.float32))
        lv = conf.lattice_vectors()
        self.assertTrue((lv[0] == N.array([1., 2., 4.], N.float32)).all())
        self.assertTrue((lv[1] == N.array([8., 4., 2.], N.float32)).all())
        self.assertTrue((lv[2] == N.array([16., 4., 8.], N.float32)).all())
        self.assertAlmostEqual(conf.cell_volume(), 168.)

def suite():
    loader = unittest.TestLoader()
    s = unittest.TestSuite()
    s.addTest(loader.loadTestsFromTestCase(AtomDescriptorTest))
    s.addTest(loader.loadTestsFromTestCase(WaterTest))
    s.addTest(loader.loadTestsFromTestCase(PeptideTest))
    s.addTest(loader.loadTestsFromTestCase(ErrorCheckingTest))
    s.addTest(loader.loadTestsFromTestCase(UniverseTest))
    s.addTest(loader.loadTestsFromTestCase(PBCTest))
    return s

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python

#-----------------------------------------------------------------------------
#       Copyright (C) 2013 The Mosaic Development Team
#
#  Distributed under the terms of the BSD License.  The full license is in
#  the file LICENSE.txt, distributed as part of this software.
#-----------------------------------------------------------------------------

from setuptools import setup, Command
import os
import sys

package_dir = "lib"
script_dir = "scripts"
test_dir = "tests"
root_dir = os.path.split(os.path.join(os.getcwd(), sys.argv[0]))[0]
commands = {}

##################################################
# Run unit tests

from unittest import TextTestRunner

class TestCommand(Command):
    """Runs the unit tests"""

    user_options = []
    def initialize_options(self):
        pass
    def finalize_options(self):
        pass
    def run(self):
        sys.path.insert(0, os.path.join(root_dir, package_dir))
        sys.path.insert(0, os.path.join(root_dir, test_dir))
        import all_tests
        t = TextTestRunner(verbosity=2)
        t.run(all_tests.suite())

commands['test'] = TestCommand

##################################################
# Build documentation if sphinx is installed

try:

    from sphinx.setup_command import BuildDoc as SphinxBuildDoc

    class BuildDoc(SphinxBuildDoc):
        def run(self):
            # Make sure the python path is pointing to the newly built
            # code so that the documentation is built on this and not a
            # previously installed version.
            build = self.get_finalized_command('build')
            sys.path.insert(0, os.path.abspath(build.build_lib))
            SphinxBuildDoc.run(self)
            sys.path.pop(0)

    commands['build_sphinx'] = BuildDoc

except ImportError:
    pass

##################################################


with open('README.txt') as file:
    long_description = file.read()

class Dummy:
    pass
version = Dummy()
exec(open('lib/mosaic/version.py').read(), version.__dict__)

setup(name='pyMosaic',
      version=version.version,
      description='MOlecular SimulAtion Interchange Conventions',
      long_description=long_description,
      author='Konrad Hinsen',
      author_email='research@khinsen.fastmail.net',
      url='http://github.com/mosaic-data-model/mosaic-python',
      license='BSD',
      package_dir = {'': package_dir},
      packages=['mosaic', 'mosaic_pdb'],
      scripts=[os.path.join(script_dir, s) for s in os.listdir(script_dir)],
      platforms=['any'],
      requires=["numpy (>=1.6)", "h5py (>=2.1)", "ImmutablePy (>=0.1)"],
      provides=["pyMosaic"],
      classifiers=[
          "Development Status :: 3 - Alpha",
          "Intended Audience :: Science/Research",
          "License :: OSI Approved :: BSD License",
          "Operating System :: OS Independent",
          "Programming Language :: Python :: 2.7",
          "Programming Language :: Python :: 3.2",
          "Programming Language :: Python :: 3.3",
          "Topic :: Scientific/Engineering",
      ],
      cmdclass = commands,
     )



# -*- coding: utf-8 -*-

#-----------------------------------------------------------------------------
#       Copyright (C) 2013 The Mosaic Development Team
#
#  Distributed under the terms of the BSD License.  The full license is in
#  the file LICENSE.txt, distributed as part of this software.
#-----------------------------------------------------------------------------

import string
import sys
import weakref

import numpy as N


# Python 2/3 compatibility issues
if sys.version_info[0] == 2:

    def ascii(string):
        return string

    def py_str(byte_string):
        if isinstance(byte_string, N.ndarray):
            return str(byte_string)
        else:
            assert isinstance(byte_string, str)
            return byte_string

    def isstring(s):
        return isinstance(s, basestring)

    xml_encoding = "utf-8"

else:

    def ascii(string):
        return bytes(string, 'ASCII')

    def py_str(byte_string):
        if isinstance(byte_string, N.ndarray):
            byte_string = bytes(byte_string)
        assert isinstance(byte_string, bytes)
        return byte_string.decode('ASCII')

    def isstring(s):
        return isinstance(s, str)

    xml_encoding = "unicode"


# Test for ASCII string without control characters.
# This doesn't look very efficient, but it works under
# both Python 2 and Python 3. Under Python 2, it refuses
# objects of type unicode.
ascii_chars = string.ascii_letters + string.digits + string.punctuation + ' '


def isascii(s):
    return isinstance(s, str) and all(c in ascii_chars for c in s)


def uint_for_max_value(m):
    if m < 1 << 8:
        return N.uint8
    elif m < 1 << 16:
        return N.uint16
    elif m < 1 << 32:
        return N.uint32
    else:
        return N.uint64


class SymbolDict(dict):

    def __getitem__(self, item):
        i = dict.get(self, item, None)
        if i is None:
            i = len(self)
            self[item] = i
        return i


class MethodRegister(object):

    def __init__(self):
        self.register_cls = {}
        self.register_str = {}

    def __call__(self, type_id):
        def register(method):
            if isinstance(type_id, str):
                self.register_str[type_id] = method
            else:
                self.register_cls[type_id] = method
            return method
        return register

    def __getitem__(self, type_id):
        if isinstance(type_id, str):
            return self.register_str[type_id]
        else:
            for t, m in self.register_cls.items():
                if issubclass(type_id, t):
                    return m
            return None


class AbstractFactory(object):

    def __init__(self):
        self.memo = weakref.WeakKeyDictionary()

    def __call__(self, obj):
        try:
            return self.memo[obj]
        except KeyError:
            h = self.handler[type(obj)]
            if h is None:
                raise TypeError("No factory routine for %s"
                                % str(type(obj)))
            new_obj = h(self, obj)
            self.memo[obj] = new_obj
            return new_obj



# -*- coding: utf-8 -*-

"""mmCIF parser

The functions in this module parse mmCIF files and return the contained
data items. There are four different routines depending on the level of
access that the application requires.

.. moduleauthor:: Konrad Hinsen <konrad.hinsen@cnrs-orleans.fr>

"""

#-----------------------------------------------------------------------------
#       Copyright (C) 2013 The Mosaic Development Team
#
#  Distributed under the terms of the BSD License.  The full license is in
#  the file LICENSE.txt, distributed as part of this software.
#-----------------------------------------------------------------------------

# The first stage of this parser (tokenizer) was inspired by the mmCIF parser
# in PyMMLIB (http://pymmlib.sourceforge.net/). The regular expression for
# identifying tokens was taken directly from there. The higher-level part of
# the parser is completely different, however. The PyMMLIB parser loads
# all the file contents into memory. This turned out to consume too much
# memory for large structure factor file. The parser in this module stores
# nothing at all. It is written as an iterator yielding mmCIF data items
# to the calling routine. The highest-level interface provides a filter
# for selecting data items and feeds tables to the calling program line
# by line through a callback object. This ensures that only the required
# data is stored, and directly in the form that the application will use.

import re
import gzip
import os
import sys

import numpy as np

# Python 2/3 compatibility issues
if sys.version_info[0] == 2:
    from urllib import urlretrieve
    bytes2text = lambda x: x
else:
    from urllib.request import urlretrieve
    from io import TextIOWrapper
    def bytes2text(stream):
        return TextIOWrapper(stream, encoding="utf8")

# Identifiers used in the return values of the parsers.
TOKEN = 'token'
DATA_LABEL = 'data_label'
DATA_VALUE = 'data_value'
KEYWORD = 'keyword'

LABEL_OR_KEYWORD = 'label_or_keyword'
VALUE = 'value'
LOOP_LABELS = 'loop_labels'
LOOP_VALUES = 'loop_values'

DATA = 'data'
TABLE_HEADER = 'table_header'
TABLE_DATA = 'table_data'

# A regular expression that encodes mmCIF syntax.
_token_regexp = re.compile(
    r"(?:"
     "(?:_(.+?)[.](\S+))"               "|"  # _section.subsection
     "(?:['\"](.*?)(?:['\"]\s|['\"]$))" "|"  # quoted strings
     "(?:\s*#.*$)"                      "|"  # comments
     "(\S+)"                                 # unquoted tokens
     ")")


# mmCIF-specific Exception classes

class MMCIFError(Exception):
    pass


class MMCIFSyntaxError(MMCIFError):

    def __init__(self, text, line_number=None):
        if line_number is None:
            # This is a workaround for the __setstate__ method in Exception
            # that effectively requires exceptions to accept a single
            # argument, otherwise a bug occurs during unpickling.
            MMCIFError.__init__(self, text)
            return
        self.line_number = line_number
        MMCIFError.__init__(self, "Line %d: %s" % (line_number, text))


#
# The parser object. It provides interfaces at different levels.
#
class MMCIFParser(object):

    """
    Parser for mmCIF files

    The file to be parsed is specified when the parser object is created.
    One of the parser functions/generators may then be called to access
    the file contents.
    """

    def __init__(self, file_name=None, file_object=None, pdb_code=None):
        """
        Specify the mmCIF file to be loaded. Only one of the four
        keyword parameters may be given a value.

        :param file_name: the name of a file
        :type file_name: str
        :param file_object: a file object
        :type file_object: file
        :param pdb_code: the PDB code for a structure file, which is
            taken from a public or local PDB repository
        :type pdb_code: str
        """
        self.line_number = 0
        if file_name is not None:
            assert file_object is None
            assert pdb_code is None
            self.file_object = open(file_name)
        elif file_object is not None:
            assert pdb_code is None
            self.file_object = file_object
        elif pdb_code is not None:
            self.file_object = mmcif_files.getFile(pdb_code)
        else:
            raise ValueError("No input file given")

    def parseLowLevel(self):
        """
        An iterator that yields the contents of the mmCIF file in the
        form of (type, data) pairs. The type can be KEYWORD, DATA_LABEL,
        or DATA_VALUE.
        """
        file_iter = iter(self.file_object)
        while True:
            line = next(file_iter)
            self.line_number += 1

            ## skip comments
            if line.startswith("#"):
                continue

            ## semi-colon multi-line strings
            if line.startswith(";"):
                lmerge = [line[1:]]
                while True:
                    line = next(file_iter)
                    self.line_number += 1
                    if line.startswith(";"):
                        break
                    lmerge.append(line)

                lmerge[-1] = lmerge[-1].rstrip()
                yield (DATA_VALUE, "".join(lmerge))
                continue

            ## split line into tokens
            for match in _token_regexp.finditer(line):
                label1, label2, string, token = match.groups()
                if label1 is not None and label2 is not None:
                    yield DATA_LABEL, (label1, label2)
                elif string is not None:
                    yield DATA_VALUE, string
                elif token is not None:
                    token_parts = token.split('_')
                    if len(token_parts) == 1 \
                       or token_parts[0].lower() not in ("data", "loop",
                                                         "global", "save",
                                                         "stop"):
                        yield DATA_VALUE, token
                    else:
                        yield KEYWORD, (token_parts[0].lower(),
                                        '_'.join(token_parts[1:]))

    def parse(self):
        """
        An iterator that yields the contents of the mmCIF file in the
        form of (type, data) pairs. The type can be KEYWORD (data
        is an optional label), DATA (data is a triple of (category label,
        item label, value)), TABLE_HEADER (data is a list of the category and
        item labels in the table) or TABLE_DATA (data is a list containing the
        data items for one row of the table).
        """
        iterator = self.parseLowLevel()
        while True:
            item_type, item = next(iterator)
            if item_type is KEYWORD and item[0] == "data":
                yield item_type, item
                break

        state = LABEL_OR_KEYWORD
        while True:
            item_type, item = next(iterator)

            if state is LABEL_OR_KEYWORD:
                if item_type is DATA_LABEL:
                    label1, label2 = item
                    state = VALUE
                elif item_type is KEYWORD:
                    if item[0] == "loop":
                        loop_labels = []
                        state = LOOP_LABELS
                    else:
                        yield item_type, item
                else:
                    raise MMCIFSyntaxError("Expected data label or keyword",
                                           self.line_number)

            elif state is VALUE:
                if item_type is DATA_VALUE:
                    state = LABEL_OR_KEYWORD
                    yield DATA, (label1, label2, item)
                else:
                    raise MMCIFSyntaxError("Expected data value "
                                           "for label %s.%s"
                                           % (label1, label2),
                                           self.line_number)

            elif state is LOOP_LABELS:
                if item_type is DATA_LABEL:
                    if loop_labels and loop_labels[0][0] != item[0]:
                        # The label does not belong to the loop category.
                        # meaning that the loop is empty and terminated.
                        label1, label2 = item
                        state = VALUE
                    else:
                        loop_labels.append(item)
                elif item_type is DATA_VALUE:
                    loop_data = [item]
                    state = LOOP_VALUES
                    yield TABLE_HEADER, loop_labels
                    if len(loop_labels) == 1:
                        yield TABLE_DATA, loop_data
                        loop_data = []
                else:
                    raise MMCIFSyntaxError("Expected label or value in loop",
                                           self.line_number)

            elif state is LOOP_VALUES:
                if item_type is DATA_VALUE:
                    loop_data.append(item)
                    if len(loop_data) == len(loop_labels):
                        yield TABLE_DATA, loop_data
                        loop_data = []
                else:
                    if len(loop_data) > 0:
                        raise MMCIFSyntaxError("Extraneous data in loop:" +
                                               str(loop_data),
                                               self.line_number)
                    if item_type is DATA_LABEL:
                        label1, label2 = item
                        state = VALUE
                    elif item_type is KEYWORD:
                        if item[0] == "loop":
                            loop_labels = []
                            state = LOOP_LABELS
                        else:
                            yield item_type, item
                    else:
                        raise MMCIFSyntaxError("Expected data label or loop",
                                               self.line_number)

    def parseAndSelect(self, categories, data=0):
        """
        An iterator that yields a subset of the contents of the mmCIF file
        in the form of (type, data) pairs. The return values are the same
        as for parse. However, only items corresponding to the selection
        are yielded. The selection consists of a list of categories and of
        the specification of a data set by number (0 = first) or name.
        Note that most mmCIF files contain only a single data set.

        :param categories: a sequence of category names
        :type categories: sequence of str
        :param data: the selected data set, either by number (0 being the
            first data set) or by name
        :type data: int or str
        """
        dataset = -1
        dataset_name = None
        for item_type, item in self.parse():

            if item_type is KEYWORD:
                if item[0] == 'data':
                    dataset += 1
                    dataset_name = item[1]
                    return_data = data == dataset or data == dataset_name
                else:
                    raise MMCIFError("Keyword %s not yet implemented"
                                     % item[0])

            elif item_type is DATA:
                if item[0] in categories and return_data:
                    yield DATA, item

            elif item_type is TABLE_HEADER:
                keep_table = False
                for label1, label2 in item:
                    if label1 in categories:
                        keep_table = True
                if keep_table and return_data:
                    yield TABLE_HEADER, item

            elif item_type is TABLE_DATA:
                if keep_table and return_data:
                    yield TABLE_DATA, item

            else:
                raise MMCIFSyntaxError("Unexpected item type %s"
                                       % str(item_type),
                                       self.line_number)

    def parseToObjects(self, data=0, **categories):
        """
        Parse the file and store the selected data using the provided
        keyword parameters. Each keyword argument specifies one category
        of data items to be processed; categories for which there is no
        keyword parameter are ignored. The value of the keyword argument
        should be a dictionary for standard data items (which are stored
        in the dictionary) and a callable object for tables. The object
        is called once for each row in the table. The two arguments
        given are 1) a dictionary mapping item labels to indices into
        the data list and 2) the items in the row as a list.

        :param data: the selected data set, either by number (0 being the
            first data set) or by name
        :type data: int or str
        :param categories: the category-specific handlers
        :type categories: dict
        """
        accumulator = None
        for item_type, item in self.parseAndSelect(categories.keys(), data):

            if item_type is DATA:
                label1, label2, value = item
                handler = categories[label1]
                if isinstance(handler, dict):
                    if accumulator is not None:
                        accumulator.deliver()
                        accumulator = None
                    handler[label2] = value
                else:
                    if accumulator is not None \
                           and not accumulator.sameLabel(label1):
                        accumulator.deliver()
                        accumulator = None
                    if accumulator is None:
                        accumulator = DataAccumulator(label1, handler)
                    accumulator.add(label2, value)

            else:

                if accumulator is not None:
                    accumulator.deliver()
                    accumulator = None

                if item_type is TABLE_HEADER:
                    indices = {}
                    for i, (label1, label2) in enumerate(item):
                        indices[label2] = i
                    handler = categories[label1]

                elif item_type is TABLE_DATA:
                    handler(indices, item)

                else:
                    raise MMCIFSyntaxError("Unexpected item type %s"
                                           % str(item_type),
                                           self.line_number)


class DataAccumulator(object):

    def __init__(self, label, handler):
        self.label = label
        self.handler = handler
        self.fields = []
        self.values = []

    def add(self, field, value):
        self.fields.append(field)
        self.values.append(value)

    def deliver(self):
        self.handler(dict((f, i) for i, f in  enumerate(self.fields)),
                     self.values)

    def sameLabel(self, label):
        return self.label == label


#
# Repository of PDB files
#

class PDBFileCollection(object):

    """
    A PDBFileCollection describes a database directory in a local PDB
    repository, containing either PDB files, mmCIF files, or structure
    factor files.
    """

    def __init__(self, base_path, filename_pattern, url_pattern):
        """
        :param base_path: the path to the database directory, i.e. the
                          path containing the two-letter subdirectories
        :type base_path: str
        :param filename_pattern: the pattern of the file names in the
                                 database, containing %s at the place of
                                 the four-letter PDB code.
        :type filename_pattern: str
        :param url_pattern: the pattern of the URLs for download from the
                            PDB server, containing %s at the place of
                            the four-letter PDB code.
        :type url_pattern: str
        """
        self.base_path = base_path
        self.is_local = base_path is not None
        self.filename_pattern = filename_pattern
        self.url_pattern = url_pattern
        self.pdb_code_index = self.filename_pattern.find('%s')

    def getFilename(self, pdb_code):
        """
        :param pdb_code: the four-letter PDB code
        :type pdb_code: str
        :return: the corresponding file name
        :rtype: str
        """
        assert len(pdb_code) == 4, "Invalid PDB code " + repr(pdb_code)
        if self.base_path is None:
            raise IOError("Directory path undefined")
        pdb_code = pdb_code.upper()
        subdir = pdb_code[1:3]
        return os.path.join(self.base_path, subdir,
                            self.filename_pattern % pdb_code)

    def fileExists(self, pdb_code):
        """
        :param pdb_code: the four-letter PDB code
        :type pdb_code: str
        :return: True if there is a corresponding file
        :rtype: bool
        """
        return os.path.exists(self.getFilename(pdb_code))

    def getFile(self, pdb_code):
        """
        :param pdb_code: the four-letter PDB code
        :type pdb_code: str
        :return: the corresponding file
        :rtype: file
        """
        if not self.is_local:
            assert len(pdb_code) == 4, "Invalid PDB code " + repr(pdb_code)
            pdb_code = pdb_code.upper()
            if self.url_pattern is None:
                raise IOError("No URL pattern for PDB repository")
            url = self.url_pattern % pdb_code
            filename, headers = urlretrieve(url)
            if url.endswith('.gz'):
                return bytes2text(gzip.GzipFile(filename))
            else:
                return open(filename)
        filename = self.getFilename(pdb_code)
        if filename.endswith('.gz'):
            return bytes2text(gzip.GzipFile(filename))
        else:
            return open(filename)

    def __iter__(self):
        """
        :return: a generator yielding the PDB codes for all the files
                 in the collection
        :rtype: generator
        """
        for dirpath, dirname, filenames in os.walk(self.base_path):
            for filename in filenames:
                pdb_code = filename[self.pdb_code_index:
                                    self.pdb_code_index + 4]
                if self.filename_pattern % pdb_code == filename:
                    yield pdb_code


pdb_mmcif_path = os.environ.get('PDB_MMCIF_PATH', None)
mmcif_files = PDBFileCollection(pdb_mmcif_path, '%s.cif.gz',
                                'http://www.rcsb.org/pdb/files/%s.cif.gz')



# -*- coding: utf-8 -*-
"""The Mosaic Python API

.. moduleauthor:: Konrad Hinsen

This module provides abstract base classes that define the Mosaic
Python API and implement validation code. They also provide a few
convenience functions implemented in terms of the raw API.

Concrete implementations subclass the abstract base classes and
implement all the abstract properties.

"""

#-----------------------------------------------------------------------------
#       Copyright (C) 2013 The Mosaic Development Team
#
#  Distributed under the terms of the BSD License.  The full license is in
#  the file LICENSE.txt, distributed as part of this software.
#-----------------------------------------------------------------------------

from abc import ABCMeta, abstractproperty
import collections
import itertools as IT
import re

import numpy as N
import numpy.linalg as LA

import mosaic.utility


# Mosaic version number (major, minor)
# An increase in the minor number indicates a superset of the preceding
# versions. An increase in the major number indicated an incompatibility
# with preceding versions.

MOSAIC_VERSION = (1, 0)


# Base class for all classes that represent top-level data items,
# i.e. items that can be stored in files, retrieved, etc.

class MosaicDataItem(object):

    """Base class for top-level data items

    Instances of subclasses of MosaicDataItem can be stored in files.
    """

    __metaclass__ = ABCMeta


# An atom is defined by the following characteristics:
#
#  - a type, which is
#    - 'element' for a standard atom that has a chemical element.
#      The element symbol is the value of the name attribute.
#    - 'cgparticle' , for particles in coarse-grained models
#      representing multiple atoms.
#    - 'dummy' for pseudo-atoms that don't physically exist,
#      such as virtual interaction sites.
#    - '' for anything else
#
# - a name, which is the chemical element symbol for atoms of type
#   'element', and any suitable identifier for the other types
#
# - a label, which identifies the atom uniquely inside its fragment
#
# - the number of sites, equal to the number of Cartesian coordinate
#   sets required by the atom. It is > 1 for path integral, wave
#   functions, atoms with alternate positions in crystal structures, etc.

class MosaicAtom(object):

    """Atom inside a :py:class:`MosaicUniverse`

    See the :ref:`data model documentation<mosaic-atom>` for atoms.

    """

    __metaclass__ = ABCMeta

    # API properties

    @abstractproperty
    def label(self):
        """An ASCII string not containing dots and identifying
        the atom uniquely inside its parent fragment.
        See the :ref:`data model documentation<mosaic-atom-label>`.
        """
        raise NotImplementedError()

    @abstractproperty
    def name(self):
        """An ASCII string describing the type of the atom. For 'real'
        atoms in the chemical sense, this must be the chemical element
        symbol.
        See the :ref:`data model documentation<mosaic-atom-name>`.
        """
        raise NotImplementedError()

    @abstractproperty
    def type(self):
        """A string identifying the type of the atom.
        See the :ref:`data model documentation<mosaic-atom-type>`.
        """
        raise NotImplementedError()

    @abstractproperty
    def number_of_sites(self):
        """The number of sites associated with the atom.
        See the :ref:`data model documentation<mosaic-atom-nsites>`.
        """
        raise NotImplementedError()

    # Property shared by all atoms

    @property
    def number_of_atoms(self):
        """The number of atoms associated with the atom (always 1)
        """
        return 1

    # Equivalence test

    def validate_equivalence(self, other):
        """Verify the equivalence of Python objects representing
        Mosaic data. The two objects can belong to different models;
        only the common API functionality is used for the test.

        :parameter other: an arbitrary Python object
        :raises ValueError: if other is not equivalent to self
        """
        if not isinstance(other, MosaicAtom):
            raise TypeError("%s is not an atom" % str(type(other)))
        if self.label != other.label:
            raise ValueError("labels differ: %s != %s"
                             % (repr(self.label), repr(other.label)))
        if self.name != other.name:
            raise ValueError("names differ: %s != %s"
                             % (repr(self.name), repr(other.name)))
        if self.type != other.type:
            raise ValueError("types differ: %s != %s"
                             % (repr(self.type), repr(other.type)))
        if self.number_of_sites != other.number_of_sites:
            raise ValueError("site numbers differ: %d != %d"
                             % (self.number_of_sites, other.number_of_sites))

    def is_equivalent(self, other):
        """Check for equivalence of Python objects representing
        Mosaic data. The two objects can belong to different models;
        only the common API functionality is used for the test.

        :parameter other: an arbitrary Python object
        :returns: True if other is equivalent to self
        :rtype: bool
        """
        try:
            self.validate_equivalence(other)
            return True
        except:
            return False

    # Validation

    _allowed_types = ('element', 'cgparticle', 'dummy', '')

    _elements = ('Ac', 'Ag', 'Al', 'Am', 'Ar', 'As', 'At', 'Au',
                 'B', 'Ba', 'Be', 'Bh', 'Bi', 'Bk', 'Br',
                 'C', 'Ca', 'Cd', 'Ce', 'Cf', 'Cl', 'Cm',
                 'Co', 'Cn', 'Cr', 'Cs', 'Cu',
                 'D', 'Db', 'Ds', 'Dy',
                 'Er', 'Es', 'Eu',
                 'F', 'Fe', 'Fm', 'Fr',
                 'Ga', 'Gd', 'Ge',
                 'H', 'He', 'Hf', 'Hg', 'Ho', 'Hs',
                 'I', 'In', 'Ir',
                 'K', 'Kr',
                 'La', 'Li', 'Lr', 'Lu',
                 'Md', 'Mg', 'Mn', 'Mo', 'Mt',
                 'N', 'Na', 'Nb', 'Nd', 'Ne', 'Ni', 'No', 'Np',
                 'O', 'Os',
                 'P', 'Pa', 'Pb', 'Pd', 'Pm', 'Po', 'Pr', 'Pt', 'Pu',
                 'Ra', 'Rb', 'Re', 'Rf', 'Rg', 'Rh', 'Rn', 'Ru',
                 'S', 'Sb', 'Sc', 'Se', 'Sg', 'Si', 'Sm', 'Sn', 'Sr',
                 'Ta', 'Tb', 'Tc', 'Te', 'Th', 'Ti', 'Tl', 'Tm',
                 'U', 'V', 'W', 'Xe', 'Y', 'Yb', 'Zn', 'Zr')

    @classmethod
    def validate_element_name(self, name):
        validate_value(name, self._elements, "name")

    def validate(self):
        """Verify that the object satisfies the constraints of the
        Mosaic data model.

        :raises ValueError: if the object is not valid Mosaic data
        """
        validate_label(self.label, "Atom.label")
        validate_value(self.type, self._allowed_types, "Atom type")
        validate_label(self.name, "Atom.name")
        if self.type == 'element':
            validate_value(self.name, self._elements, "Atom element")
        if not isinstance(self.number_of_sites, int):
            raise ValueError("Atom.number_of_sites must be an integer")
        if self.number_of_sites <= 0:
            raise ValueError("Atom.number_of_sites must be positive")
        if self.number_of_atoms != 1:
            raise ValueError("Atom.number_of_atoms must be 1")


# A fragment describes a node in the tree defining the chemical
# structure of the molecules. It can represent a molecule,
# a supermolecule, or part of a molecule. A fragment is defined
# by
#
#  - a species, which is a text string describing the chemical
#    nature of the fragment
#
#  - a label, which describes the role of the fragment inside its
#    parent structure, and must be unique within that parent structure
#
#  - a list of sub-fragments
#
#  - a list of atoms
#
#  - a list of bonds
#
#  - the boolean flag is_polymer. Polymer fragments have an emtpy
#    atom list and an additional attribute 'polymer_type' whose
#    allowed values are
#      - 'polypeptide', for a peptide chain
#      - 'polyribonucleotide', for an RNA chain
#      - 'polydeoxyribonucleotide', for a DNA chain
#      - 'polynucleotide', for a chain that can contain
#        nucleotides with either type of sugar
#      - the empty string, for any other polymer

class MosaicFragment(collections.Mapping):

    """Fragment inside a :py:class:`MosaicUniverse`

    See the :ref:`data model documentation<mosaic-fragment>` for fragments.

    Fragments implement the Mapping interface. Valid keys are strings
    identifying atoms or sub-fragments, using dots to separate subsequent
    labels. For polymer fragments, integer keys are valid as well to refer
    to a specific sub-fragment in the chain.
    """

    # API properties

    @abstractproperty
    def label(self):
        """An ASCII string not containing dots and identifying
        the fragment uniquely inside its parent fragment.
        See the :ref:`data model documentation<mosaic-fragment-label>`.
        """
        raise NotImplementedError()

    @abstractproperty
    def species(self):
        """An ASCII string describing the species of the fragment
        (e.g. what kind of molecule or moiety it represents).
        See the :ref:`data model documentation<mosaic-fragment-species>`.
        """
        raise NotImplementedError()

    @abstractproperty
    def fragments(self):
        """Sequence of sub-fragments, may be empty.
        See the :ref:`data model documentation<mosaic-fragment-fragments>`.
        """
        raise NotImplementedError()

    @abstractproperty
    def atoms(self):
        """Sequence of atoms, may be empty.
        See the :ref:`data model documentation<mosaic-fragment-atoms>`.
        """
        raise NotImplementedError()

    @abstractproperty
    def bonds(self):
        """Sequence of bonds, may be empty. Each bond is
        repreented by a tuple (atom_ref_1, atom_ref_2, bond_order).
        See the :ref:`data model documentation<mosaic-fragment-bonds>`
        and the :ref:`bond reference documentation<mosaic-bonds>`.
        """
        raise NotImplementedError()

    @abstractproperty
    def is_polymer(self):
        """True if the fragment is a polymer.
        See the :ref:`data model documentation<mosaic-fragment-polymer>`.
        """
        raise NotImplementedError()

    @abstractproperty
    def polymer_type(self):
        """String identifying the polymer type if is_polymer is true.
        See the :ref:`data model documentation<mosaic-fragment-polymer>`.
        """
        raise NotImplementedError()

    # Properties that can be computed in terms of the API properties

    @property
    def number_of_atoms(self):
        """The number of atoms in the fragment (including the atoms
        in sub-fragments).
        """
        return sum(f.number_of_atoms for f in self.fragments) \
               + len(self.atoms)

    @property
    def number_of_sites(self):
        """The number of sites associated with the fragment, i.e.
        the sum of the numbers of sites of all the fragment's atoms,
        including those of sub-fragments.
        """
        return sum(f.number_of_sites for f in self.fragments) \
               + sum(a.number_of_sites for a in self.atoms)

    @property
    def number_of_bonds(self):
        """The number of bonds associated with the fragment,
        including the bonds inside sub-fragments.
        """
        return sum(ff.number_of_bonds for ff in self.fragments) \
               + len(self.bonds)

    # Equivalence test

    def validate_equivalence(self, other):
        """Verify the equivalence of Python objects representing
        Mosaic data. The two objects can belong to different models;
        only the common API functionality is used for the test.

        :parameter other: an arbitrary Python object
        :raises ValueError: if other is not equivalent to self
        """
        if not isinstance(other, MosaicFragment):
            raise TypeError("%s is not a fragment" % str(type(other)))
        if self.label != other.label:
            raise ValueError("labels differ: %s != %s"
                             % (repr(self.label), repr(other.label)))
        if self.species != other.species:
            raise ValueError("species differ: %s != %s"
                             % (repr(self.species), repr(other.species)))
        if self.is_polymer:
            if not other.is_polymer:
                raise ValueError("%s is not a polymer" % str(other))
            if self.polymer_type != other.polymer_type:
                raise ValueError("polymer types differ: %s != %s"
                                 % (repr(self.polymer_type),
                                    repr(other.polymer_type)))
        for s, o in zip(self.fragments, other.fragments):
            s.validate_equivalence(o)
        for s, o in zip(self.atoms, other.atoms):
            s.validate_equivalence(o)
        if self._bond_set() != other._bond_set():
            raise ValueError("bonds differ: %s != %s"
                             % (repr(self._bond_set), repr(other._bond_set)))

    def _bond_set(self):
        return frozenset((frozenset((a1, a2)), order)
                         for a1, a2, order in self.bonds)

    def is_equivalent(self, other):
        """Check for equivalence of Python objects representing
        Mosaic data. The two objects can belong to different models;
        only the common API functionality is used for the test.

        :parameter other: an arbitrary Python object
        :returns: True if other is equivalent to self
        :rtype: bool
        """
        try:
            self.validate_equivalence(other)
            return True
        except:
            return False

    # Validation

    _polymer_types = ['',
                      'polypeptide',
                      'polyribonucleotide',
                      'polydeoxyribonucleotide',
                      'polynucleotide']

    _bond_orders = ['', 'single', 'double', 'triple', 'quadruple', 'aromatic']

    def validate(self):
        """Verify that the object satisfies the constraints of the
        Mosaic data model.

        :raises ValueError: if the object is not valid Mosaic data
        """
        validate_label(self.label, "Fragment.label")
        validate_label(self.species, "Fragment.species")
        validate_value(self.is_polymer, [True, False], "Fragment.is_polymer")
        if self.is_polymer:
            validate_value(self.polymer_type, self._polymer_types,
                           "Fragment.polymer_type")
        validate_sequence(self.fragments, MosaicFragment, "Fragment.fragments")
        labels = set()
        for f in self.fragments:
            f.validate()
            if f.label in labels:
                raise ValueError("Label %s occurs more than once" % f.label)
            labels.add(f.label)
        validate_sequence(self.atoms, MosaicAtom, "Fragment.atoms")
        for a in self.atoms:
            a.validate()
            if a.label in labels:
                raise ValueError("Label %s occurs more than once" % a.label)
            labels.add(a.label)
        for property in ['number_of_atoms',
                         'number_of_sites',
                         'number_of_bonds']:
            value = getattr(self, property)
            reference = getattr(MosaicFragment, property).fget(self)
            if value != reference:
                raise ValueError("Fragment.%s is %s, should be %s"
                                 % (property, str(value), str(reference)))

    # Methods based on API properties

    def recursive_atom_iterator(self):
        """An iterator over the atoms in the fragment, including the
        atoms in sub-fragments.
        """
        return IT.chain(IT.chain.from_iterable(f.recursive_atom_iterator()
                                               for f in self.fragments),
                        iter(self.atoms))

    def recursive_atom_path_iterator(self):
        """An iterator over the atom paths in the fragment, including the
        atoms in sub-fragments.
        """
        for f in self.fragments:
            l = f.label
            for ap in f.recursive_atom_path_iterator():
                yield l + '.' + ap
        for a in self.atoms:
            yield a.label

    def recursive_bond_iterator(self):
        """An iterator over the bonds in the fragment, including the
        bonds in sub-fragments.
        """
        for f in self.fragments:
            l = f.label
            for a1, a2, order in f.recursive_bond_iterator():
                yield (l + '.' + a1, l + '.' + a2, order)
        for b in self.bonds:
            yield b

    def site_to_atom_index_mapping(self):
        """
        :returns: an array whose element [s] is the atom index
                  corresponding to site index s.
        :rtype: numpy.ndarray
        """
        ns = [a.number_of_sites for a in self.recursive_atom_iterator()]
        return N.repeat(N.arange(len(ns)), ns)

    def _atom_to_site_index_mapping(self):
        ns = [a.number_of_sites for a in self.recursive_atom_iterator()]
        return N.add.accumulate(ns)

    def atom_to_site_index_mapping(self):
        """
        :returns: an array whose element [s] is the site index
                  corresponding to the first site of the atom with
                  atom index s.
        :rtype: numpy.ndarray
        """
        m = self._atom_to_site_index_mapping()
        m[1:] = m[:-1]
        m[0] = 0
        return m

    # Mapping interface

    def __getitem__(self, item):
        if isinstance(item, int) and self.is_polymer:
            return self.fragments[item]
        # A rather inefficient implementation of substructure selection
        # using only the public API elements. Concrete implementations
        # can do better.
        assert isinstance(item, str)
        path = item.split('.')
        item = None
        for f in self.fragments:
            if f.label == path[0]:
                item = f
                break
        if item is None:
            for a in self.atoms:
                if a.label == path[0]:
                    item = a
                    break
        if item is None:
            raise KeyError(path[0])
        if len(path) > 1:
            return item['.'.join(path[1:])]
        else:
            return item

    def __len__(self):
        """
        :returns: the number of sub-elements, i.e. atoms and sub-fragments
        :rtype: int
        """
        return len(self.fragments) + len(self.atoms)

    def __iter__(self):
        """Iterate over sub-fragments first, then over the fragment's atoms.
        """
        for f in self.fragments:
            yield f.label
        for a in self.atoms:
            yield a.label

    # Override some methods from collections.Mapping for efficiency

    def values(self):
        return self.fragments + self.atoms

    def itervalues(self):
        return IT.chain(iter(self.fragments), iter(self.atoms))


# A universe description consists of
#
# - the cell shape
#
# - a list of symmetry transformations
#
# - the chemical structure of its contents
#
# - a label indicating additional conventions that this universe
#   conforms to, in particular naming conventions for atoms and molecules

class MosaicUniverse(MosaicDataItem):

    """Universe

    See the :ref:`data model documentation<mosaic-universe>` for universes.

    """

    # API properties

    @abstractproperty
    def cell_shape(self):
        """A string identifying the cell shape.
        See the :ref:`data model documentation<mosaic-universe-cell-shape>`.
        """
        raise NotImplementedError()

    @abstractproperty
    def symmetry_transformations(self):
        """A sequence of symmetry transformations, possibly empty.
        Each symmetry tranformation is defined by a two-element tuple,
        whose first item is a rotation matrix and whose second item
        is a translation vector.
        See the :ref:`data model documentation<mosaic-universe-symmetry>`.
        """
        raise NotImplementedError()

    @abstractproperty
    def convention(self):
        """An ASCII string naming the conventions used inside the
        definitions of fragements and atoms.
        See the :ref:`data model documentation<mosaic-universe-convention>`.
        """
        raise NotImplementedError()

    @abstractproperty
    def molecules(self):
        """A sequence of molecule specifications. Each element is a
        two-element tuple, whose first element is a :py:class:`MosaicFragment`
        and whose second element is an integer specifying the number of
        copies of the molecule.
        See the :ref:`data model documentation<mosaic-universe-molecules>`.
        """
        raise NotImplementedError()

    # Properties that can be computed in terms of the API properties

    _cell_parameter_array_shapes = {
        'infinite': (0,),
        'cube': (),
        'cuboid': (3,),
        'parallelepiped': (3, 3)}

    @property
    def cell_parameter_array_shape(self):
        """The shape of a valid cell_parameters array
        in a :py:class:`MosaicConfiguration`.
        """
        return self._cell_parameter_array_shapes[self.cell_shape]

    @property
    def number_of_molecules(self):
        "The number of molecules in the universe."
        return sum(n for f, n in self.molecules)

    @property
    def number_of_atoms(self):
        "The number of atoms in the universe."
        return sum(n * f.number_of_atoms for f, n in self.molecules)

    @property
    def number_of_sites(self):
        "The number of sites in the universe."
        return sum(n * f.number_of_sites for f, n in self.molecules)

    @property
    def number_of_bonds(self):
        "The number of bonds in the universe."
        return sum(n * f.number_of_bonds for f, n in self.molecules)

    @property
    def number_of_template_atoms(self):
        """The number of template atoms in the universe, i.e. the
        total number of atoms in all fragment definitions. It is
        equal to the number of atoms iff all molecule repetition
        counts are 1.
        """
        return sum(f.number_of_atoms for f, n in self.molecules)

    @property
    def number_of_template_sites(self):
        """The number of template sites in the universe, i.e. the
        total number of sites in all fragment definitions. It is
        equal to the number of sites iff all molecule repetition
        counts are 1.
        """
        return sum(f.number_of_sites for f, n in self.molecules)

    # Equivalence test

    def validate_equivalence(self, other):
        """Verify the equivalence of Python objects representing
        Mosaic data. The two objects can belong to different models;
        only the common API functionality is used for the test.

        :parameter other: an arbitrary Python object
        :raises ValueError: if other is not equivalent to self
        """
        if not isinstance(other, MosaicUniverse):
            raise TypeError("%s is not a universe" % str(type(other)))
        if self.cell_shape != other.cell_shape:
            raise ValueError("cell shapes differ: %s != %s"
                             % (repr(self.cell_shape), repr(other.cell_shape)))
        if self.convention != other.convention:
            raise ValueError("naming conventions differ: %s != %s"
                             % (repr(self.convention),
                                repr(other.convention)))
        for (r1, t1), (r2, t2) in zip(self.symmetry_transformations,
                                      other.symmetry_transformations):
            if (r1 != r2).any():
                raise ValueError("rotation matrices differ: %s != %s"
                                 % (str(r1), str(r2)))
            if (t1 != t2).any():
                raise ValueError("translation vectors differ: %s != %s"
                                 % (str(t1), str(t2)))
        for (sf, sc), (of, oc) in zip(self.molecules, other.molecules):
            sf.validate_equivalence(of)
            if sc != oc:
                raise ValueError("molecule counts differ: %d != %d" % (sc, oc))

    def is_equivalent(self, other):
        """Check for equivalence of Python objects representing
        Mosaic data. The two objects can belong to different models;
        only the common API functionality is used for the test.

        :parameter other: an arbitrary Python object
        :returns: True if other is equivalent to self
        :rtype: bool
        """
        try:
            self.validate_equivalence(other)
            return True
        except:
            return False

    # Validation

    def validate(self):
        """Verify that the object satisfies the constraints of the
        Mosaic data model.

        :raises ValueError: if the object is not valid Mosaic data
        """
        validate_value(self.cell_shape,
                       self._cell_parameter_array_shapes.keys(),
                       "Universe.cell_shape")
        validate_label(self.convention, "Universe.convention")
        validate_sequence(self.symmetry_transformations,
                          collections.Sequence,
                          "Universe.symmetry_transformations",
                          ((lambda t: len(t) == 2,
                            "must have length 2"),
                           (lambda rot, trans:
                            getattr(rot, "shape", None) == (3, 3),
                            "rotation matrix shape is not (3, 3)"),
                           (lambda rot, trans:
                            getattr(trans, "shape", None) == (3,),
                            "translation vector shape is not (3,)"),
                           (lambda rot, trans:
                            rot.dtype == N.float64
                            and trans.dtype == N.float64,
                            "rotation and translation must be float64")))
        if self.cell_shape == "infinite" \
           and len(self.symmetry_transformations) > 0:
            raise ValueError("Symmetry transformations are allowed "
                             "only in periodic universes")
        for f, n in self.molecules:
            f.validate()
            if not isinstance(n, int):
                raise ValueError("Molecule count must be an integer")
            if n <= 0:
                raise ValueError("Molecule count must be positive")
        for property in ['cell_parameter_array_shape',
                         'number_of_molecules', 'number_of_atoms',
                         'number_of_sites', 'number_of_bonds',
                         'number_of_template_atoms',
                         'number_of_template_sites']:
            value = getattr(self, property)
            reference = getattr(MosaicUniverse, property).fget(self)
            if value != reference:
                raise ValueError("Universe.%s is %s, should be %s"
                                 % (property, str(value), str(reference)))

    # Methods based on API properties

    def recursive_atom_iterator(self):
        """An iterator over the atoms in the universe.
        """
        for fragment, count in self.molecules:
            for _ in range(count):
                for a in fragment.recursive_atom_iterator():
                    yield a

    def bond_index_array(self):
        """Returns an integer array of shape (N, 2), where N
        is the total number of bonds in the universe. The entries
        [i, 0] and [i, 1] refer to the two atoms that are connected
        by bond i. The entry [i, 0] is smaller than the entry [i, 1].

        :returns: the bond index array
        :rtype: numpy.ndarray
        """
        natoms = 0
        bonds = []
        for fragment, count in self.molecules:
            f_paths = list(fragment.recursive_atom_path_iterator())
            f_bonds = [sorted((f_paths.index(a1), f_paths.index(a2)))
                       for a1, a2, order in fragment.recursive_bond_iterator()]
            for _ in range(count):
                bonds.extend([(natoms+a1, natoms+a2) for a1, a2 in f_bonds])
                natoms += len(f_paths)
        return N.array(bonds)

    def site_to_atom_index_mapping(self):
        """
        :returns: an array whose element [s] is the atom index
                  corresponding to site index s.
        :rtype: numpy.ndarray
        """
        natoms = 0
        total = []
        for fragment, count in self.molecules:
            per_fragment = fragment.site_to_atom_index_mapping()
            f_natoms = fragment.number_of_atoms
            for _ in range(count):
                total.append(per_fragment + natoms)
                natoms += f_natoms
        return N.concatenate(total)

    def atom_to_site_index_mapping(self):
        """
        :returns: an array whose element [s] is the site index
                  corresponding to the first site of the atom with
                  atom index s.
        :rtype: numpy.ndarray
        """
        nsites = 0
        total = []
        for fragment, count in self.molecules:
            per_fragment = fragment._atom_to_site_index_mapping()
            for _ in range(count):
                total.append(per_fragment + nsites)
                nsites += per_fragment[-1]
        m = N.concatenate(total)
        m[1:] = m[:-1]
        m[0] = 0
        return m

    def template_site_to_template_atom_index_mapping(self):
        """
        :returns: an array whose element [s] is the template atom index
                  corresponding to template site index s.
        :rtype: numpy.ndarray
        """
        natoms = 0
        total = []
        for fragment, count in self.molecules:
            per_fragment = fragment.site_to_atom_index_mapping()
            f_natoms = fragment.number_of_atoms
            total.append(per_fragment + natoms)
            natoms += f_natoms
        return N.concatenate(total)

    def site_to_template_index_mapping(self):
        """
        :returns: an array whose element [s] is the template site index
                  corresponding to site index s.
        :rtype: numpy.ndarray
        """
        ntsites = 0
        total = []
        for fragment, count in self.molecules:
            f_nsites = fragment.number_of_sites
            per_fragment = N.arange(f_nsites)
            for _ in range(count):
                total.append(per_fragment + ntsites)
            ntsites += f_nsites
        return N.concatenate(total)

    def atom_to_template_index_mapping(self):
        """
        :returns: an array whose element [s] is the template atom index
                  corresponding to atom index s.
        :rtype: numpy.ndarray
        """
        ntatoms = 0
        total = []
        for fragment, count in self.molecules:
            f_natoms = fragment.number_of_atoms
            per_fragment = N.arange(f_natoms)
            for _ in range(count):
                total.append(per_fragment + ntatoms)
            ntatoms += f_natoms
        return N.concatenate(total)


# Properties associate a value with each atom or site in a
# universe. The value can be an array of any shape and any element
# type, but shape and element type are the same for all atoms.
# Properties also have a name that describes the quantitity they
# store, and the units of this quantity.
#
# Properties whose type starts with "template" are defined only for
# each atom or site in the molecule templates, not for each individual
# molecular instance. They are used for properties that are the same
# for all molecules of the same type (e.g. atomic mass).

class MosaicProperty(MosaicDataItem):

    """Property

    See the :ref:`data model documentation<mosaic-property>` for properties.

    """

    # API properties

    @abstractproperty
    def type(self):
        """A string identifying the type of the property.
        See the :ref:`data model documentation<mosaic-property-type>`.
        """
        raise NotImplementedError()

    @abstractproperty
    def name(self):
        """An ASCII string describing the property.
        See the :ref:`data model documentation<mosaic-property-name>`.
        """
        raise NotImplementedError()

    @abstractproperty
    def units(self):
        """A string identifying the physical units of the property.
        See the :ref:`data model documentation<mosaic-property-units>`.
        """
        raise NotImplementedError()

    @abstractproperty
    def universe(self):
        "The :py:class:`MosaicUniverse` for which the property is defined."
        raise NotImplementedError()

    @abstractproperty
    def data(self):
        """An array containing the property's values.
        See the :ref:`data model documentation<mosaic-property-data>`.
        """
        raise NotImplementedError()

    # Properties that can be computed in terms of the API properties

    @property
    def element_shape(self):
        """The shape of the sub-array containing the propery for one
        atom or site.
        """
        return self.data.shape[1:]

    # Equivalence test

    def validate_equivalence(self, other):
        """Verify the equivalence of Python objects representing
        Mosaic data. The two objects can belong to different models;
        only the common API functionality is used for the test.

        :parameter other: an arbitrary Python object
        :raises ValueError: if other is not equivalent to self
        """
        if not isinstance(other, MosaicProperty):
            raise TypeError("%s is not a property" % str(type(other)))
        self.universe.validate_equivalence(other.universe)
        if self.type != other.type:
            raise ValueError("types differ: %s != %s"
                             % (repr(self.type), repr(other.type)))
        if self.name != other.name:
            raise ValueError("names differ: %s != %s"
                             % (repr(self.name), repr(other.name)))
        if self.units != other.units:
            raise ValueError("units differ: %s != %s"
                             % (repr(self.units), repr(other.units)))
        if self.element_shape != other.element_shape:
            raise ValueError("element shapes differ: %s != %s"
                             % (repr(self.element_shape),
                                repr(other.element_shape)))
        if self.data.dtype != other.data.dtype:
            raise ValueError("data dtypes differ: %s != %s"
                             % (repr(self.data.dtype),
                                repr(other.data.dtype)))
        if (self.data != other.data).any():
            raise ValueError("data arrays differ")

    def is_equivalent(self, other):
        """Check for equivalence of Python objects representing
        Mosaic data. The two objects can belong to different models;
        only the common API functionality is used for the test.

        :parameter other: an arbitrary Python object
        :returns: True if other is equivalent to self
        :rtype: bool
        """
        try:
            self.validate_equivalence(other)
            return True
        except:
            return False

    # Validation

    _allowed_dtypes = [N.int8, N.int16, N.int32, N.int64,
                       N.uint8, N.uint16, N.uint32, N.uint64,
                       N.float32, N.float64,
                       N.bool]

    _allowed_types = ["atom", "site", "template_atom", "template_site"]

    def validate(self):
        """Verify that the object satisfies the constraints of the
        Mosaic data model.

        :raises ValueError: if the object is not valid Mosaic data
        """
        validate_value(self.type, self._allowed_types, "Property.type")
        validate_label(self.name, "Property.name")
        validate_units(self.units, "Property.units")
        validate_type(self.universe, MosaicUniverse, "Property.universe")
        self.universe.validate()
        el_shape = self.element_shape
        data_shape = ({"atom": self.universe.number_of_atoms,
                       "site": self.universe.number_of_sites,
                       "template_atom": self.universe.number_of_template_atoms,
                       "template_site": self.universe.number_of_template_sites,
                       }[self.type],) + el_shape
        validate_array(self.data,
                       data_shape,
                       self._allowed_dtypes,
                       "Property.data")


# Labels associate a text string with each atom or site in a
# universe. They work much like properties, except for having a string
# value. Labels are a separate data item because the differences
# compared to numerical properties (no element shape, no unit) would
# make validation of a common data item type too complicated.
#
# Labels whose type starts with "template" are defined only for
# each atom or site in the molecule templates, not for each individual
# molecular instance.

class MosaicLabel(MosaicDataItem):

    """Label

    See the :ref:`data model documentation<mosaic-label>` for labels.

    """

    # API properties

    @abstractproperty
    def type(self):
        """A string identifying the type of the label.
        See the :ref:`data model documentation<mosaic-label-type>`.
        """
        raise NotImplementedError()

    @abstractproperty
    def name(self):
        """An ASCII string describing the label.
        See the :ref:`data model documentation<mosaic-label-name>`.
        """
        raise NotImplementedError()

    @abstractproperty
    def universe(self):
        "The :py:class:`MosaicUniverse` for which the property is defined."
        raise NotImplementedError()

    @abstractproperty
    def strings(self):
        """A sequence of strings representing a label for each atom or site.
        See the :ref:`data model documentation<mosaic-label-strings>`.
        """
        raise NotImplementedError()

    # Equivalence test

    def validate_equivalence(self, other):
        """Verify the equivalence of Python objects representing
        Mosaic data. The two objects can belong to different models;
        only the common API functionality is used for the test.

        :parameter other: an arbitrary Python object
        :raises ValueError: if other is not equivalent to self
        """
        if not isinstance(other, MosaicLabel):
            raise TypeError("%s is not a label" % str(type(other)))
        self.universe.validate_equivalence(other.universe)
        if self.type != other.type:
            raise ValueError("types differ: %s != %s"
                             % (repr(self.type), repr(other.type)))
        if self.name != other.name:
            raise ValueError("names differ: %s != %s"
                             % (repr(self.name), repr(other.name)))
        for s1, s2 in zip(self.strings, other.strings):
            if s1 != s2:
                raise ValueError("labels differ: %s != %s"
                                 % (repr(s1), repr(s2)))

    def is_equivalent(self, other):
        """Check for equivalence of Python objects representing
        Mosaic data. The two objects can belong to different models;
        only the common API functionality is used for the test.

        :parameter other: an arbitrary Python object
        :returns: True if other is equivalent to self
        :rtype: bool
        """
        try:
            self.validate_equivalence(other)
            return True
        except:
            return False

    # Validation

    _allowed_types = ["atom", "site", "template_atom", "template_site"]

    def validate(self):
        """Verify that the object satisfies the constraints of the
        Mosaic data model.

        :raises ValueError: if the object is not valid Mosaic data
        """
        validate_value(self.type, self._allowed_types, "Label.type")
        validate_label(self.name, "Label.name")
        validate_type(self.universe, MosaicUniverse, "Label.universe")
        self.universe.validate()
        validate_sequence(self.strings, str, "Label.strings")
        nstrings = {"atom": self.universe.number_of_atoms,
                    "site": self.universe.number_of_sites,
                    "template_atom": self.universe.number_of_template_atoms,
                    "template_site": self.universe.number_of_template_sites,
                    }[self.type]
        if len(self.strings) != nstrings:
            raise ValueError("incorrect number of strings")
        for s in self.strings:
            validate_ascii_string(s, "label")


# A configuration specifies a coordinate for each site and
# the shape and size of the cell.

class MosaicConfiguration(MosaicDataItem):

    """Configuration

    See the :ref:`data model documentation<mosaic-configuration>`
    for configurations.

    """

    # API properties

    @abstractproperty
    def universe(self):
        "The :py:class:`MosaicUniverse` for which the property is defined."
        raise NotImplementedError()

    @abstractproperty
    def cell_parameters(self):
        """An array containing the parameters defining the shape and size
        of the cell.
        See the :ref:`data model documentation<mosaic-configuration-cp>`.
        """
        raise NotImplementedError()

    @abstractproperty
    def positions(self):
        """An array containing an (x, y, z) position for each site.
        See the :ref:`data model documentation<mosaic-configuration-pos>`.
        """
        raise NotImplementedError()

    # Methods based on API properties

    def lattice_vectors(self):
        """
        :returns: a sequence of arrays of shape (3,) containing the lattice
                  vectors for the simulation cell.
        :rtype: tuple
        """
        return {'infinite': lambda p: (),
                'cube': lambda p: (N.array([p, 0., 0.], dtype=p.dtype),
                                   N.array([0., p, 0.], dtype=p.dtype),
                                   N.array([0., 0., p], dtype=p.dtype)),
                'cuboid': lambda p: (N.array([p[0], 0., 0.], dtype=p.dtype),
                                     N.array([0., p[1], 0.], dtype=p.dtype),
                                     N.array([0., 0., p[2]], dtype=p.dtype)),
                'parallelepiped': lambda p: tuple(N.array(v) for v in p),
                }[self.universe.cell_shape](self.cell_parameters)

    def cell_volume(self):
        """
        :returns: the volume the simulation cell, or None for
                  infinite universes
        :rtype: float
        """
        return {'infinite': lambda p: None,
                'cube': lambda p: float(p*p*p),
                'cuboid': lambda p: p[0]*p[1]*p[2],
                'parallelepiped': lambda p: abs(LA.det(p)),
                }[self.universe.cell_shape](self.cell_parameters)

    # Equivalence test

    def validate_equivalence(self, other):
        """Verify the equivalence of Python objects representing
        Mosaic data. The two objects can belong to different models;
        only the common API functionality is used for the test.

        :parameter other: an arbitrary Python object
        :raises ValueError: if other is not equivalent to self
        """
        if not isinstance(other, MosaicConfiguration):
            raise TypeError("%s is not a configuration" % str(type(other)))
        self.universe.validate_equivalence(other.universe)
        if (self.cell_parameters != other.cell_parameters).any():
            raise ValueError("cell parameters differ: %s != %s"
                             % (repr(self.cell_parameters),
                                repr(other.cell_parameters)))
        if self.cell_parameters.dtype != other.cell_parameters.dtype:
            raise ValueError("cell parameter dtypes differ: %s != %s"
                             % (repr(self.cell_parameters.dtype),
                                repr(other.cell_parameters.dtype)))
        if self.positions.dtype != other.positions.dtype:
            raise ValueError("position dtypes differ: %s != %s"
                             % (repr(self.positions.dtype),
                                repr(other.positions.dtype)))
        if (self.positions != other.positions).any():
            raise ValueError("position arrays differ")

    def is_equivalent(self, other):
        """Check for equivalence of Python objects representing
        Mosaic data. The two objects can belong to different models;
        only the common API functionality is used for the test.

        :parameter other: an arbitrary Python object
        :returns: True if other is equivalent to self
        :rtype: bool
        """
        try:
            self.validate_equivalence(other)
            return True
        except:
            return False

    # Validation

    _allowed_dtypes = [N.float32, N.float64]

    def validate(self):
        """Verify that the object satisfies the constraints of the
        Mosaic data model.

        :raises ValueError: if the object is not valid Mosaic data
        """
        validate_type(self.universe, MosaicUniverse, "Configuration.universe")
        self.universe.validate()
        validate_array(self.cell_parameters,
                       self.universe.cell_parameter_array_shape,
                       self._allowed_dtypes,
                       "Configuration.cell_parameters")
        validate_array(self.positions,
                       (self.universe.number_of_sites, 3),
                       self._allowed_dtypes,
                       "Configuration.positions")
        if self.cell_parameters.dtype != self.positions.dtype:
            raise ValueError("Configuration.cell_parameters and "
                             "Configuration.positions must have same dtypes")


# Selections specify a subset of atoms or sites, either in the templates
# or in the molecules of a universe.

class MosaicSelection(MosaicDataItem):

    """Selection

    See the :ref:`data model documentation<mosaic-selection>` for selections.

    """

    # API properties

    @abstractproperty
    def type(self):
        """A string identifying the type of the selection.
        See the :ref:`data model documentation<mosaic-selection-type>`.
        """
        raise NotImplementedError()

    @abstractproperty
    def universe(self):
        "The :py:class:`MosaicUniverse` for which the selection is defined."
        raise NotImplementedError()

    @abstractproperty
    def indices(self):
        """An array containing the indices of the contained atoms or sites.
        See the :ref:`data model documentation<mosaic-selection-indices>`.
        """
        raise NotImplementedError()

    # Properties that can be computed in terms of the API properties

    @property
    def number_of_atoms(self):
        """The number of atoms in the selection.
        """
        indices = self.indices
        if self.type == "atom":
            return len(indices)
        elif self.type == "template_atom":
            natoms = 0
            ntatoms = 0
            for fragment, count in self.universe.molecules:
                nta = fragment.number_of_atoms
                natoms += count * N.sum((indices >= ntatoms)
                                        & (indices < ntatoms+nta))
                ntatoms += nta
            return natoms
        else:
            raise TypeError("number of atoms undefined in site selection")

    @property
    def number_of_sites(self):
        """The number of sites in the selection.
        """
        indices = self.indices
        if self.type == "atom":
            sites_per_atom = [a.number_of_sites
                              for a in self.universe.recursive_atom_iterator()]
            # If 'indices' is an immutable array, N.take crashes due to
            # a numpy bug. Conversion to a plain array prevents this.
            # See Github issue #3758 for numpy/numpy.
            return N.sum(N.take(sites_per_atom, N.array(indices)))
        elif self.type == "template_atom":
            sites_per_atom = [a.number_of_sites
                              for a in self.universe.recursive_atom_iterator()]
            nsites = 0
            ntatoms = 0
            for fragment, count in self.universe.molecules:
                nta = fragment.number_of_atoms
                mask = (indices >= ntatoms) & (indices < ntatoms+nta)
                ntatoms += nta
                # Conversion to plain arrays is required for the same
                # reason as above.
                nsites += count * N.sum(N.take(sites_per_atom,
                                               N.repeat(N.array(indices),
                                                        N.array(mask))))
            return nsites
        elif self.type == "site":
            return len(indices)
        elif self.type == "template_site":
            nsites = 0
            ntsites = 0
            for fragment, count in self.universe.molecules:
                nts = fragment.number_of_sites
                nsites += count * N.sum((indices >= ntsites)
                                        & (indices < ntsites+nts))
                ntsites += nts
            return nsites

    # Equivalence test

    def validate_equivalence(self, other):
        """Verify the equivalence of Python objects representing
        Mosaic data. The two objects can belong to different models;
        only the common API functionality is used for the test.

        :parameter other: an arbitrary Python object
        :raises ValueError: if other is not equivalent to self
        """
        if not isinstance(other, MosaicSelection):
            raise TypeError("%s is not a selection" % str(type(other)))
        self.universe.validate_equivalence(other.universe)
        if self.type != other.type:
            raise ValueError("types differ: %s != %s"
                             % (repr(self.type), repr(other.type)))
        if (self.indices != other.indices).any():
            raise ValueError("indices differ")

    def is_equivalent(self, other):
        """Check for equivalence of Python objects representing
        Mosaic data. The two objects can belong to different models;
        only the common API functionality is used for the test.

        :parameter other: an arbitrary Python object
        :returns: True if other is equivalent to self
        :rtype: bool
        """
        try:
            self.validate_equivalence(other)
            return True
        except:
            return False

    # Validation

    _allowed_types = ["atom", "site", "template_atom", "template_site"]

    def validate(self):
        """Verify that the object satisfies the constraints of the
        Mosaic data model.

        :raises ValueError: if the object is not valid Mosaic data
        """
        validate_value(self.type, self._allowed_types, "Selection.type")
        validate_type(self.universe, MosaicUniverse, "Selection.universe")
        max_index = {"atom": self.universe.number_of_atoms,
                     "site": self.universe.number_of_sites,
                     "template_atom": self.universe.number_of_template_atoms,
                     "template_site": self.universe.number_of_template_sites,
                    }[self.type]
        validate_indices(self.indices, max_index, "Selection.indices")

# Validation functions

def validate_type(obj, cls, text):
    if isinstance(obj, cls):
        return
    raise TypeError("%s must be of type %s (is %s)"
                    % (text, cls.__name__, str(type(obj))))


def validate_value(value, allowed_values, name):
    if not value in allowed_values:
        raise ValueError("%s must be one of %s"
                         % (name, ", ".join(allowed_values)))


def validate_ascii_string(s, text):
    if not mosaic.utility.isascii(s):
        raise ValueError("non-ASCII string in %s" % text)


def validate_label(label, text):
    validate_type(label, str, text)
    if len(label) > 32767:
        raise ValueError("%s too long, must be <= 32767 characters" % text)
    for c in label:
        if not c in _allowed_in_labels:
            raise ValueError("illegal character '%s' in %s"
                             % (c, text))

_allowed_in_labels = 'abcdefghijklmnopqrstuvwxyz' + \
                     'ABCDEFGHIJKLMNOPQRSTUVWXYZ' + \
                     '0123456789!#$%&?@^_~+-*/=,()[]' + "'"


def validate_array(array, shape, allowed_dtypes, text):
    # Don't check for N.ndarray in order to allow on-disk
    # arrays (HDF5, netCDF)
    try:
        a_shape = array.shape
        a_dtype = array.dtype
    except AttributeError:
        raise TypeError("%s must be an array" % text)
    if shape is not None and a_shape != shape:
        raise ValueError("%s must have shape %s"
                         % (text, str(shape)))
    if a_dtype not in allowed_dtypes:
        raise TypeError(" %s must have element type %s"
                        % (text,
                           " or ".join(str(t) for t in allowed_dtypes)))


def validate_sequence(obj, el_cls, text, additional_tests=()):
    if not (isinstance(obj, collections.Sequence) and
            all(isinstance(item, el_cls) for item in obj)):
        raise TypeError("%s must be a sequence of %s elements"
                        % (text, el_cls.__name__))
    for test_fn, text in additional_tests:
        for el in obj:
            if not test_fn(el):
                raise ValueError("%s: %s" % (str(el), text))

def validate_indices(indices, max_index, text):
    validate_array(indices, None,
                   [N.uint8, N.uint16, N.uint32, N.uint64],
                   text)
    if len(indices.shape) != 1:
        raise ValueError("index array not 1d")
    if (indices >= max_index).any():
        raise ValueError("index too large")
    if len(indices) > 1:
        d = indices[1:]-indices[:-1]
        if (d <= 0).any():
            raise ValueError("indices not sorted")

# The unit validator accepts a very limited syntax, which
# should be sufficient: a unit is defined by a string of
# unit names with an optional numeric suffix indication a power.
# The unit list can include integers or decimal fractions.
#
# Examples: "nm ps-1" (velocity), "nm2" (area), "kJ mol-1" (energy)
#           "0.1 nm" (length), "60 s" (time)

def validate_units(unit_spec, text):
    validate_type(unit_spec, str, text)
    first = True
    for unit in unit_spec.split():
        # number (integer or decimal fraction)
        if first:
            first = False
            if (re.match("^[1-9][0-9]*$", unit) \
                or re.match("^0\.[0-9]+$", unit)):
                continue
        # symbol+exponent
        m = re.match("^(?P<symbol>[a-zA-Z]+)([-]?[0-9]+)?$", unit)
        if m and m.group('symbol') in _allowed_units:
            continue
        raise ValueError("invalid unit '%s' in %s" % (unit, unit_spec))

_allowed_units = \
    ["pm",   "Ang",  "nm",   "um",   "mm",   "m",    
     "fs",   "ps",   "ns",   "us",   "ms",   "s",    
     "amu",  "g",    "kg",   
     "mol",  
     "J",    "kJ",   "cal",  "kcal", "eV",   
     "K",    
     "Pa",   "kPa",  "MPa",  "GPa",  "atm",  "bar",  "kbar", 
     "e",    "C",    "A",    "V",    
     "rad",  
     "c",    "h",    "me"]

# Validation as a test function rather than raising exceptions

def is_valid(obj):
    try:
        obj.validate()
        return True
    except:
        return False
